class Member:
    """Represents a Weverse Member.

    .. container:: operations

        .. describe:: x == y

            Checks if two members are equal.

        .. describe:: x != y

            Checks if two members are not equal.

        .. describe:: hash(x)

            Returns the member's hash.

        .. describe:: str(x)

            Returns the member's name.

    Attributes
    ----------
    data: :class:`str`
        The raw data directly taken from the response generated by Weverse's API.
    id: :class:`str`
        The ID of the member.
    community_id: :class:`int`
        The community ID of the community the member belongs to. This
        can be used to fetch the actual :class:`sunverse.objects.Community`
        object if needed using the :method:`fetch_community` method.
    joined: :class:`bool`
        Whether the member has joined the community.
    name: :class:`str`
        The name of the member.
    image_url: Optional[:class:`str`]
        The URL of the profile image of the member, if any.
    banner_url: Optional[:class:`str`]
        The URL of the profile banner of the member, if any.
    profile_comment: Optional[:class:`str`]
        The profile comment of the member, if any.
    follower_count: :class:`int`
        The number of users following the member.
    member_join_status: :class:`str`
        The join status of the member.
    profile_type: :class:`str`
        The profile type of the member.
    hidden: :class:`bool`
        UNDETERMINED FUNCTIONALITY
    blinded: :class:`bool`
        UNDETERMINED FUNCTIONALITY
    has_membership :class:`bool`
        Whether the member has a paid subscription to the community.
    has_official_mark: :class:`bool`
        Whether the member is an official Weverse account.
    joined_date: :class:`int`
        The date the member joined the community on, in epoch.
    """

    __slots__ = (
        "data",
        "id",
        "community_id",
        "joined",
        "name",
        "image_url",
        "banner_url",
        "profile_comment",
        "follower_count",
        "member_join_status",
        "profile_type",
        "hidden",
        "blinded",
        "has_membership",
        "has_official_mark",
        "joined_date",
    )

    def __init__(self, data: dict):
        self.data: dict = data
        self.id: str = data["memberId"]
        self.community_id: int = data["communityId"]
        self.joined: bool = data["joined"]
        self.name: str = data["profileName"]
        self.image_url: str | None = data.get("profileImageUrl")
        self.banner_url: str | None = data.get("profileCoverImageUrl")
        self.profile_comment: str | None = data.get("profileComment")
        self.follower_count: int = (
            data["followCount"]["followerCount"] if data.get("followCount") else None
        )
        self.member_join_status: str = data["memberJoinStatus"]
        self.profile_type: str = data["profileType"]
        self.hidden: bool = data["hidden"]
        self.blinded: bool = data["blinded"]
        self.has_membership: bool = data["hasMembership"]
        self.has_official_mark: bool = data["hasOfficialMark"]
        self.joined_date: int = data["joinedDate"]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __repr__(self):
        return f"Member member_id={self.id}, name={self.name}"

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.id)



from typing import Optional

from sunverse.objects.attachments import Photo
from sunverse.objects.post import Postlike


class Medialike(Postlike):
    """Represents a Weverse Media.

    Inherits from :class:`objects.post.Postlike`.

    Shares the same attributes with :class:`objects.post.Postlike`.

    Inherited by:

    - :class:`objects.media.ImageMedia`
    - :class:`objects.media.YoutubeMedia`
    - :class:`objects.media.WeverseMedia`
    - :class:`objects.media.Live`

    .. container:: operations

        .. describe:: str(x)

            Returns the media's title.

    Attributes
    ----------
    title: :class:`str`
        The title of the media.
    thumbnail_url: :class:`str`
        The URL of the thumbnail of the media.
    """

    __slots__ = ("title", "thumbnail_url")

    def __init__(self, data: dict):
        super().__init__(data)
        self.title: str = data["title"]
        self.thumbnail_url: str = data["extension"]["mediaInfo"]["thumbnail"]["url"]

    def __repr__(self):
        return f"Media media_id={self.id}, title={self.title}"

    def __str__(self):
        return self.title


class ImageMedia(Medialike):
    """Represents a Weverse Media that contains images.

    Inherits from :class:`sunverse.objects.Medialike`.

    Shares the same attributes with :class:`sunverse.objects.Postlike` and
    :class:`sunverse.objects.Medialike`.
    """

    @property
    def photos(self) -> list[Photo]:
        """list[:class:`sunverse.objects.Photo`]: A list of
        :class:`sunverse.objects.Photo` objects in the media.
        """
        return [
            Photo(photo_data)
            for photo_data in self.data["extension"]["image"]["photos"]
        ]


class YoutubeMedia(Medialike):
    """Represents a Weverse Media that contains a YouTube video.

    Inherits from :class:`sunverse.objects.Medialike`.

    Shares the same attributes with :class:`sunverse.objects.Postlike` and
    :class:`sunverse.objects.Medialike`.

    Attributes
    ----------
    video_duration: :class:`int`
        The duration of the YouTube video, in seconds.
    youtube_url: :class:`str`
        The URL to the YouTube video.
    video_screen_orientation: :class:`str`
        The screen orientation of the video.
    """

    __slots__ = ("video_duration", "youtube_url", "video_screen_orientation")

    def __init__(self, data: dict):
        super().__init__(data)
        self.video_duration: int = data["extension"]["youtube"]["playTime"]
        self.youtube_url: str = data["extension"]["youtube"]["videoPath"]
        self.video_screen_orientation: str = data["extension"]["youtube"][
            "screenOrientation"
        ]


class WeverseMedia(Medialike):
    """Represents a Weverse Media that contains a Weverse video.

    Inherits from :class:`sunverse.objects.Medialike`.

    Shares the same attributes with :class:`sunverse.objects.Postlike` and
    :class:`sunverse.objects.Medialike`.

    Inherited by:

    - :class:`sunverse.objects.Live`

    Attributes
    ----------
    video_id: :class:`int`
        The ID of the Weverse video.
    internal_video_id: Optional[:class:`str`]
        The internal ID of the Weverse video. Returns `None`
        in the case of a live broadcast that is still broadcasting and not
        converted into a VOD yet.
    video_type: :class:`str`
        The type of Weverse video.
    aired_at: :class:`int`
        The time the Weverse video got aired at, in epoch time.
    paid_video: :class:`bool`
        Whether the Weverse video is a paid video.
    membership_only_video: :class:`bool`
        Whether the Weverse video is only accessible by users with a membership.
    video_screen_orientation: :class:`str`
        The screen orientation of the video.
    video_play_count: :class:`int`
        The number of views on the Weverse video.
    video_like_count: :class:`int`
        The number of likes on the Weverse video.
    video_duration: Optional[:class:`int`]
        The duration of the Weverse video, in seconds. Returns `None`
        in the case of a live broadcast that is still broadcasting and not
        converted into a VOD yet.
    """

    __slots__ = (
        "video_id",
        "internal_video_id",
        "video_type",
        "aired_at",
        "paid_video",
        "membership_only_video",
        "video_screen_orientation",
        "video_play_count",
        "video_like_count",
        "video_duration",
    )

    def __init__(self, data: dict):
        super().__init__(data)
        self.video_id: int = data["extension"]["video"]["videoId"]
        self.internal_video_id: Optional[str] = data["extension"]["video"].get(
            "infraVideoId"
        )
        self.video_type: str = data["extension"]["video"]["type"]
        self.aired_at: int = data["extension"]["video"]["onAirStartAt"]
        self.paid_video: bool = data["extension"]["video"]["paid"]
        self.membership_only_video: bool = data["extension"]["video"]["membershipOnly"]
        self.video_screen_orientation: str = data["extension"]["video"][
            "screenOrientation"
        ]
        self.video_play_count: int = data["extension"]["video"]["playCount"]
        self.video_like_count: int = data["extension"]["video"]["likeCount"]
        self.video_duration: Optional[int] = data["extension"]["video"].get("playTime")



from typing import Any


class _MissingSentinel:
    __slots__ = ()

    def __eq__(self, other):
        return False

    def __bool__(self):
        return False

    def __hash__(self):
        return 0

    def __repr__(self):
        return "..."


MISSING: Any = _MissingSentinel()



class SunverseException(Exception):
    """Base exception class that other Sunverse exceptions inherit from."""


class LoginError(SunverseException):
    """An Exception raised when the login fails."""

    def __init__(self, code: int, reason: str):
        message = f"Login to Weverse has failed.\nCode: {code}\nReason: {reason}"
        super().__init__(message)


class RequestFailed(SunverseException):
    """An Exception raised when the response returned by the API is not 200."""

    def __init__(self, url: str, code: int, reason: str):
        message = f"Request to {url} has failed.\nCode: {code}\nReason: {reason}"
        super().__init__(message)


class NotFound(SunverseException):
    """An Exception raised when the response returned by the API is 404."""

    def __init__(self, url: str):
        message = (
            f"Request to {url} has failed.\nCode: 404\n"
            "Reason: The requested resource cannot be found or does not exist."
        )
        super().__init__(message)


class InternalServerError(SunverseException):
    """An Exception raised when the response returned by the API is 500."""

    def __init__(self, url: str):
        message = (
            f"Request to {url} has failed.\nCode: 500\n\n"
            "Reason: The Weverse API encountered an Internal Server Error."
        )
        super().__init__(message)



import re
from typing import Optional

from sunverse.enums import PostTypes


class Notification:
    """Represents a Weverse Notification.

    .. container:: operations

        .. describe:: x == y

            Checks if two notifications are equal.

        .. describe:: x != y

            Checks if two notifications are not equal.

        .. describe:: hash(x)

            Returns the notification's hash.

        .. describe:: str(x)

            Returns the notification's message.

    Attributes
    ----------
    data: :class:`dict`
        The raw data directly taken from the response generated by Weverse's API.
    id: :class:`int`
        The ID of the notification.
    is_read: :class:`bool`
        Whether the user has read the notification.
    title: :class:`str`
        Usually the name of the group the notification belongs to.
        However, if it's an admin notification, it would be a proper title.
    message_ko: :class:`str`
        The message of the notification in Korean.
    message_ja: :class:`str`
        The message of the notification in Japanese.
    message_en: :class:`str`
        The message of the notification in English.
    logo_image_url: :class:`str`
        The URL to the logo image of the artist profile
        picture displayed on the notification.
    image_url: Optional[:class:`str`]
        The URL to the preview image displayed on the notification, if any.
    web_url: :class:`str`
        The URL that leads to the actual post.
    time_created: :class:`int`
        The time the notification got created at, in epoch.
    count: :class:`int`
        The number of artist comments on the post the notification
        leads to.
    community_id: :class:`int`
        The community ID of the community the notification belongs to.
        This can be used to fetch the actual :class:`sunverse.objects.Community`
        object if needed using the :method:`fetch_community` method.
    """

    __slots__ = (
        "data",
        "id",
        "is_read",
        "title",
        "message_ko",
        "message_ja",
        "message_en",
        "logo_image_url",
        "image_url",
        "web_url",
        "time_created",
        "count",
        "community_id",
    )

    def __init__(self, data: dict):
        self.data: dict = data
        self.id: int = data["activityId"]
        self.is_read: bool = data["read"]
        self.title: str = data["title"]
        self.message_ko: str = data["message"]["values"]["ko"]
        self.message_ja: str = data["message"]["values"]["ja"]
        self.message_en: str = data["message"]["values"]["en"]
        self.logo_image_url: str = data["logoImageUrl"]
        self.image_url: Optional[str] = data.get("imageUrl")
        self.web_url: str = data["webUrl"]
        self.time_created: int = data["time"]
        self.count: int = data["count"]
        self.community_id: int = data["community"]["communityId"]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __repr__(self):
        return f"Notification notification_id={self.id}, message={self.message_en}"

    def __str__(self):
        return self.message_en

    def __hash__(self):
        return hash(self.id)

    @property
    def post_id(self) -> str:
        """:class:`str`: Returns the post ID of the post the notification leads to."""
        pattern = re.compile(r"([\d]-\d+)")
        match = re.search(pattern, self.data["messageId"])

        if not match:  # This happens when the notification type is Notice.
            match = re.search(r"\d+", self.data["messageId"])

        return match.group(0)

    @property
    def post_type(self) -> str:
        """:class:`str`: Returns the post type of the post the notification
        leads to.
        """
        post_types = {
            "T_FEED_COMMENT": PostTypes.USER_POST_COMMENT,
            "ST_FEED_COMMENT": PostTypes.USER_POST_COMMENT,
            "_MEDIA_COMMENT": PostTypes.MEDIA_COMMENT,
            "_MOMENT_COMMENT": PostTypes.MOMENT_COMMENT,
            "MOMENT_COMMENT": PostTypes.MOMENT_COMMENT,
            "ARTIST_COMMENT": PostTypes.ARTIST_POST_COMMENT,
            "ARTIST_POST": PostTypes.POST,
            "ARTIST_MOMENT": PostTypes.MOMENT,
            "ARTIST_LIVE_ON_AIR": PostTypes.LIVE,
            "COMMUNITY_MEDIA": PostTypes.MEDIA,
            "NOTICE": PostTypes.NOTICE,
            "COMMUNITY_ANNIVERSARY": PostTypes.BIRTHDAY,
        }

        for key, value in post_types.items():
            if key in self.data["messageId"]:
                return value

        return "NOT IMPLEMENTED"

    @property
    def author_id(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns the artist ID of the artist who
        created the post the notification leads to.

        Will always return `None` for posts that are not
        comment-related. This can be used together with the
        :method:`fetch_artists` method to fetch the dictionary that contains
        the :class:`sunverse.objects.Artist` objects, and then accessing the
        specific artist object using this author ID as the key.
        """
        if self.data.get("authors"):
            return self.data["authors"][0]["memberId"]

        return None



import asyncio
import json

import pytest
from sunverse.enums import ObjectTypes, PostTypes
from sunverse.objects.artist import Artist
from sunverse.objects.community import Community
from sunverse.objects.media import ImageMedia, WeverseMedia, YoutubeMedia
from sunverse.objects.notification import Notification
from sunverse.sunverse import SunverseClient


class TestSunverse:
    sunverse_client = SunverseClient("anson_2012@yahoo.com", "BearBear123!")

    @pytest.mark.asyncio
    async def test_fetch_joined_communities(self):
        """This test is to ensure that this API wrapper is kept up-to-date with
        the data structure provided by Weverse's API.

        Notes
        -----
        It is best recommended to test this with an account that has
        joined all the communities to ensure no special cases arise.
        This test makes 1 request to the Weverse API per community the
        account used is part of.
        """

        await self.sunverse_client._fetch_access_token_by_credentials()
        required_keys = [
            "communityId",
            "communityName",
            "communityAlias",
            "urlPath",
            "logoImage",
            "homeHeaderImage",
            "homeGradationColor",
            "hasMembershipProduct",
            "availableActions",
            "agencyProfile",
            "birthdayArtists",
        ]
        optional_keys = ["artistCode", "fandomName", "fanEventUrl", "memberCount"]
        agency_profile_keys = ["profileImageUrl", "profileName", "profileCoverImageUrl"]
        successful = True

        joined_communities = await self.sunverse_client.fetch_joined_communities()

        for joined_community in joined_communities:
            for required_key in required_keys:
                if required_key not in joined_community.data:
                    raise KeyError(
                        f"{required_key} does not exist in {joined_community.name}."
                    )

            for key in joined_community.data:
                if key not in optional_keys and key not in required_keys:
                    raise KeyError(
                        f"{key} is neither an optional key, nor a required key - "
                        "which means that it's a new key added on Weverse's side and "
                        "an update to the Community class is required."
                    )

            for agency_profile_key in agency_profile_keys:
                if agency_profile_key not in joined_community.data["agencyProfile"]:
                    raise KeyError(
                        f"{agency_profile_key} does not exist in the agencyProfile "
                        f"of {joined_community.name}."
                    )

            for c_agency_profile_key in joined_community.data["agencyProfile"]:
                if c_agency_profile_key not in agency_profile_keys:
                    raise KeyError(
                        f"{c_agency_profile_key} is not a agency profile key - "
                        "which means that it's a new key added on Weverse's side and "
                        "an update to the Community class is required."
                    )

        assert successful

    @pytest.mark.asyncio
    async def test_fetch_latest_notifications(self):
        """This test is to ensure that this API wrapper is kept up-to-date with
        the data structure provided by Weverse's API.
        """

        required_keys = [
            "activityId",
            "read",
            "messageId",
            "title",
            "message",
            "logoImageUrl",
            "webUrl",
            "time",
            "count",
            "community",
            "type",
        ]
        useless_keys = ["scheme"]
        optional_keys = ["imageUrl", "authors", "compactedMessageMap", "createDate"]
        successful = True

        latest_notifications = await self.sunverse_client.fetch_latest_notifications()

        for latest_notification in latest_notifications:
            for required_key in required_keys:
                if required_key not in latest_notification.data:
                    raise KeyError(
                        f"{required_key} does not exist in {latest_notification.id}."
                    )

            for key in latest_notification.data:
                if (
                    key not in optional_keys
                    and key not in required_keys
                    and key not in useless_keys
                ):
                    raise KeyError(
                        f"{key} is neither an optional key, nor a required key - "
                        "which means that it's a new key added on Weverse's side and "
                        "an update to the Notification class is required."
                    )

        assert successful

    '''@pytest.mark.asyncio
    async def test_posts(self):
        """This test is to ensure that this API wrapper is kept up-to-date with
        the data structure provided by Weverse's API."""
        loop = 0
        start = True
        successful = True
        notifications = await self.sunverse_client.fetch_latest_notifications()
        required_keys = [
            "extension",
            "shareUrl",
            "bookmarked",
            "locked",
            "hasProduct",
            "attachment",
            "body",
            "tags",
            "availableActions",
            "postType",
            "sectionType",
            "publishedAt",
            "hideFromArtist",
            "membershipOnly",
            "commentCount",
            "postId",
            "emotionCount",
            "plainBody",
            "author",
            "community",
        ]
        attachment_keys = ["photo", "video"]
        optional_keys = ["artistReactions", "viewerEmotionId"]

        while loop <= 100:
            if not start:
                await asyncio.sleep(0.25)
                data = await self.sunverse_client._fetcher(self.sunverse_client._urls.latest_notifications_url())
                data = await self.sunverse_client._fetcher(self.sunverse_client._urls.notification_url(int(data["paging"]["nextParams"]["after"])))
                notifications = [await self.sunverse_client._object_creator(ObjectTypes.NOTIFICATION, d, self.sunverse_client._urls.notification_url(int(data["paging"]["nextParams"]["after"]))) for d in data["data"] if d.get("community")]

            for notification in notifications:
                await asyncio.sleep(0.25)
                if notification.post_type != PostTypes.POST:
                    continue

                post = await self.sunverse_client.fetch_post(notification.post_id)

                for required_key in required_keys:
                    if required_key not in post.data:
                        raise KeyError(
                            f"Required Key: {required_key} does not exist "
                            f"in the fetched data for {post.url}"
                        )

                for key in post.data:
                    if key not in required_keys and key not in optional_keys:
                        raise KeyError(
                            f"Undefined Key: {key} does not exist in the required "
                            f"keys list and optional keys list for {post.url} - "
                            "which means that it's a new key added on Weverse's side and "
                            "an update to the Post class is required."
                        )

                for attachment_key in post.data["attachment"]:
                    if attachment_key not in attachment_keys:
                        raise KeyError(
                            f"Undefined Attachment Key: {attachment_key} does not "
                            f"exist in the attachment keys list for {post.url}"
                        )

            start = False
            loop += 1

        assert successful

    @pytest.mark.asyncio
    async def test_medias(self):
        """This test is to ensure that this API wrapper is kept up-to-date with
        the data structure provided by Weverse's API."""
        loop = 0
        start = True
        successful = True
        notifications = await self.sunverse_client.fetch_latest_notifications()
        required_keys = [
            "extension",
            "shareUrl",
            "bookmarked",
            "locked",
            "hasProduct",
            "attachment",
            "title",
            "body",
            "tags",
            "availableActions",
            "postType",
            "sectionType",
            "publishedAt",
            "hideFromArtist",
            "membershipOnly",
            "commentCount",
            "postId",
            "emotionCount",
            "plainBody",
            "author",
            "community",
        ]
        attachment_keys = ["photo", "video"]
        optional_keys = ["artistReactions", "viewerEmotionId"]

        while loop <= 100:
            if not start:
                await asyncio.sleep(0.25)
                data = await self.sunverse_client._fetcher(self.sunverse_client._urls.latest_notifications_url())
                data = await self.sunverse_client._fetcher(self.sunverse_client._urls.notification_url(int(data["paging"]["nextParams"]["after"])))
                notifications = [await self.sunverse_client._object_creator(ObjectTypes.NOTIFICATION, d, self.sunverse_client._urls.notification_url(int(data["paging"]["nextParams"]["after"]))) for d in data["data"] if d.get("community")]

            for notification in notifications:
                await asyncio.sleep(0.25)
                if notification.post_type != PostTypes.MEDIA:
                    continue

                media = await self.sunverse_client.fetch_media(notification.post_id)

                if media:
                    for required_key in required_keys:
                        if required_key not in media.data:
                            raise KeyError(
                                f"Required Key: {required_key} does not exist "
                                f"in the fetched data for {media.url}"
                            )

                    for key in media.data:
                        if key not in required_keys and key not in optional_keys:
                            raise KeyError(
                                f"Undefined Key: {key} does not exist in the required "
                                f"keys list and optional keys list for {media.url} - "
                                "which means that it's a new key added on Weverse's side and "
                                "an update to the Media class is required."
                            )

                    for attachment_key in media.data["attachment"]:
                        if attachment_key not in attachment_keys:
                            raise KeyError(
                                f"Undefined Attachment Key: {attachment_key} does not "
                                f"exist in the attachment keys list for {media.url}"
                            )

                    if isinstance(media, ImageMedia):
                        extra_required_keys = ["mediaInfo", "image"]

                        for extra_required_key in extra_required_keys:
                            if extra_required_key not in media.data["extension"]:
                                raise KeyError(
                                    f"Required Extra Key: {extra_required_key} does "
                                    f"not exist in the fetched data for {media.url}."
                                )

                        for extra_key in media.data["extension"]:
                            if extra_key not in extra_required_keys:
                                raise KeyError(
                                    f"Undefined Extra Key: {extra_key} does not exist "
                                    f"in the extra required keys list for {media.url} - "
                                    "which means that it's a new key added on Weverse's side "
                                    "and an update to the Media class is required."
                                )

                        required_media_keys = ["mediaType", "title", "body", "thumbnail", "categories"]

                        for required_media_key in required_media_keys:
                            if required_media_key not in media.data["extension"]["mediaInfo"]:
                                raise KeyError(
                                    f"Required Media Key: {required_media_key} does "
                                    f"not exist in the fetched data for {media.url}."
                                )

                        for media_key in media.data["extension"]["mediaInfo"]:
                            if media_key not in required_media_keys:
                                raise KeyError(
                                    f"Undefined Media Key: {media_key} does not exist "
                                    f"in the required media keys list for {media.url} - "
                                    "which means that it's a new key added on Weverse's side "
                                    "and an update to the Media class is required."
                                )

                        required_image_keys = ["photoCount", "photos", "viewType"]

                        for required_image_key in required_image_keys:
                            if required_image_key not in media.data["extension"]["image"]:
                                raise KeyError(
                                    f"Required Image Key: {required_image_key} does "
                                    f"not exist in the fetched data for {media.url}."
                                )

                        for image_key in media.data["extension"]["image"]:
                            if image_key not in required_image_keys:
                                raise KeyError(
                                    f"Undefined Image Key: {image_key} does not exist "
                                    f"in the required image keys list for {media.url} - "
                                    "which means that it's a new key added on Weverse's side "
                                    "and an update to the Media class is required."
                                )

                    if isinstance(media, WeverseMedia):
                        extra_required_keys = ["mediaInfo", "video"]

                        for extra_required_key in extra_required_keys:
                            if extra_required_key not in media.data["extension"]:
                                raise KeyError(
                                    f"Required Extra Key: {extra_required_key} does "
                                    f"not exist in the fetched data for {media.url}."
                                )

                        for extra_key in media.data["extension"]:
                            if extra_key not in extra_required_keys:
                                raise KeyError(
                                    f"Undefined Extra Key: {extra_key} does not exist "
                                    f"in the extra required keys list for {media.url} - "
                                    "which means that it's a new key added on Weverse's side "
                                    "and an update to the Media class is required."
                                )

                        required_media_keys = ["mediaType", "title", "body", "thumbnail", "categories"]

                        for required_media_key in required_media_keys:
                            if required_media_key not in media.data["extension"]["mediaInfo"]:
                                raise KeyError(
                                    f"Required Media Key: {required_media_key} does "
                                    f"not exist in the fetched data for {media.url}."
                                )

                        for media_key in media.data["extension"]["mediaInfo"]:
                            if media_key not in required_media_keys:
                                raise KeyError(
                                    f"Undefined Media Key: {media_key} does not exist "
                                    f"in the required media keys list for {media.url} - "
                                    "which means that it's a new key added on Weverse's side "
                                    "and an update to the Media class is required."
                                )

                        required_video_keys = [
                            "videoId",
                            "type",
                            "onAirStartAt",
                            "paid",
                            "membershipOnly",
                            "screenOrientation",
                            "thumb",
                            "previewType",
                            "playCount",
                            "likeCount",
                            "serviceName",
                            "serviceId",
                            "liveToVod",
                            "infraVideoId",
                            "playTime",
                            "encodingStatus"
                        ]

                        for required_video_key in required_video_keys:
                            if required_video_key not in media.data["extension"]["video"]:
                                raise KeyError(
                                    f"Required Video Key: {required_video_key} does "
                                    f"not exist in the fetched data for {media.url}."
                                )

                        for video_key in media.data["extension"]["video"]:
                            if video_key not in required_video_keys:
                                raise KeyError(
                                    f"Undefined Video Key: {video_key} does not exist "
                                    f"in the required video keys list for {media.url} - "
                                    "which means that it's a new key added on Weverse's side "
                                    "and an update to the Media class is required."
                                )

                    if isinstance(media, YoutubeMedia):
                        extra_required_keys = ["mediaInfo", "youtube"]

                        for extra_required_key in extra_required_keys:
                            if extra_required_key not in media.data["extension"]:
                                raise KeyError(
                                    f"Required Extra Key: {extra_required_key} does "
                                    f"not exist in the fetched data for {media.url}."
                                )

                        for extra_key in media.data["extension"]:
                            if extra_key not in extra_required_keys:
                                raise KeyError(
                                    f"Undefined Extra Key: {extra_key} does not exist "
                                    f"in the extra required keys list for {media.url} - "
                                    "which means that it's a new key added on Weverse's side "
                                    "and an update to the Media class is required."
                                )

                        required_media_keys = ["mediaType", "title", "body", "thumbnail", "categories"]

                        for required_media_key in required_media_keys:
                            if required_media_key not in media.data["extension"]["mediaInfo"]:
                                raise KeyError(
                                    f"Required Media Key: {required_media_key} does "
                                    f"not exist in the fetched data for {media.url}."
                                )

                        for media_key in media.data["extension"]["mediaInfo"]:
                            if media_key not in required_media_keys:
                                raise KeyError(
                                    f"Undefined Media Key: {media_key} does not exist "
                                    f"in the required media keys list for {media.url} - "
                                    "which means that it's a new key added on Weverse's side "
                                    "and an update to the Media class is required."
                                )

                        required_youtube_keys = [
                            "screenOrientation",
                            "playTime",
                            "videoPath",
                        ]

                        for required_youtube_key in required_youtube_keys:
                            if required_youtube_key not in media.data["extension"]["youtube"]:
                                raise KeyError(
                                    f"Required Youtube Key: {required_youtube_key} does "
                                    f"not exist in the fetched data for {media.url}."
                                )

                        for youtube_key in media.data["extension"]["youtube"]:
                            if youtube_key not in required_youtube_keys:
                                raise KeyError(
                                    f"Undefined Youtube Key: {youtube_key} does not exist "
                                    f"in the required youtube keys list for {media.url} - "
                                    "which means that it's a new key added on Weverse's side "
                                    "and an update to the Media class is required."
                                )

            start = False
            loop += 1

        assert successful

    @pytest.mark.asyncio
    async def test_lives(self):
        """This test is to ensure that this API wrapper is kept up-to-date with
        the data structure provided by Weverse's API."""
        loop = 0
        start = True
        successful = True
        notifications = await self.sunverse_client.fetch_latest_notifications()
        required_keys = [
            "extension",
            "shareUrl",
            "bookmarked",
            "locked",
            "hasProduct",
            "attachment",
            "title",
            "body",
            "tags",
            "availableActions",
            "postType",
            "sectionType",
            "publishedAt",
            "hideFromArtist",
            "membershipOnly",
            "commentCount",
            "postId",
            "emotionCount",
            "plainBody",
            "author",
            "community",
        ]
        attachment_keys = ["photo", "video"]
        optional_keys = ["artistReactions", "viewerEmotionId"]

        while loop <= 100:
            if not start:
                await asyncio.sleep(0.25)
                data = await self.sunverse_client._fetcher(self.sunverse_client._urls.latest_notifications_url())
                data = await self.sunverse_client._fetcher(self.sunverse_client._urls.notification_url(int(data["paging"]["nextParams"]["after"])))
                notifications = [await self.sunverse_client._object_creator(ObjectTypes.NOTIFICATION, d, self.sunverse_client._urls.notification_url(int(data["paging"]["nextParams"]["after"]))) for d in data["data"] if d.get("community")]

            for notification in notifications:
                await asyncio.sleep(0.25)
                if notification.post_type != PostTypes.LIVE:
                    continue

                live = await self.sunverse_client.fetch_live(notification.post_id)

                if live:
                    for required_key in required_keys:
                        if required_key not in live.data:
                            raise KeyError(
                                f"Required Key: {required_key} does not exist "
                                f"in the fetched data for {live.url}"
                            )

                    for key in live.data:
                        if key not in required_keys and key not in optional_keys:
                            raise KeyError(
                                f"Undefined Key: {key} does not exist in the required "
                                f"keys list and optional keys list for {live.url} - "
                                "which means that it's a new key added on Weverse's side and "
                                "an update to the Live class is required."
                            )

                    for attachment_key in live.data["attachment"]:
                        if attachment_key not in attachment_keys:
                            raise KeyError(
                                f"Undefined Attachment Key: {attachment_key} does not "
                                f"exist in the attachment keys list for {live.url}"
                            )

                    extra_required_keys = ["mediaInfo", "video"]

                    for extra_required_key in extra_required_keys:
                        if extra_required_key not in live.data["extension"]:
                            raise KeyError(
                                f"Required Extra Key: {extra_required_key} does "
                                f"not exist in the fetched data for {live.url}."
                            )

                    for extra_key in live.data["extension"]:
                        if extra_key not in extra_required_keys:
                            raise KeyError(
                                f"Undefined Extra Key: {extra_key} does not exist "
                                f"in the extra required keys list for {live.url} - "
                                "which means that it's a new key added on Weverse's side "
                                "and an update to the Live class is required."
                            )

                    required_media_keys = ["mediaType", "title", "body", "thumbnail", "chat"]
                    optional_media_keys = ["categories"]

                    for required_media_key in required_media_keys:
                        if required_media_key not in live.data["extension"]["mediaInfo"]:
                            raise KeyError(
                                f"Required Media Key: {required_media_key} does "
                                f"not exist in the fetched data for {live.url}."
                            )

                    for media_key in live.data["extension"]["mediaInfo"]:
                        if media_key not in required_media_keys and media_key not in optional_media_keys:
                            raise KeyError(
                                f"Undefined Media Key: {media_key} does not exist "
                                f"in the required media keys list for {live.url} - "
                                "which means that it's a new key added on Weverse's side "
                                "and an update to the Live class is required."
                            )

                    required_video_keys = [
                        "videoId",
                        "type",
                        "onAirStartAt",
                        "paid",
                        "membershipOnly",
                        "screenOrientation",
                        "thumb",
                        "previewType",
                        "playCount",
                        "likeCount",
                        "serviceName",
                        "serviceId",
                        "liveToVod",
                        "infraVideoId",
                        "playTime",
                        "encodingStatus"
                    ]

                    for required_video_key in required_video_keys:
                        if required_video_key not in live.data["extension"]["video"]:
                            raise KeyError(
                                f"Required Video Key: {required_video_key} does "
                                f"not exist in the fetched data for {live.url}."
                            )

                    for video_key in live.data["extension"]["video"]:
                        if video_key not in required_video_keys:
                            raise KeyError(
                                f"Undefined Video Key: {video_key} does not exist "
                                f"in the required video keys list for {live.url} - "
                                "which means that it's a new key added on Weverse's side "
                                "and an update to the Live class is required."
                            )

            start = False
            loop += 1

        assert successful'''

    @pytest.mark.asyncio
    async def test_notices(self):
        """This test is to ensure that this API wrapper is kept up-to-date with
        the data structure provided by Weverse's API."""
        loop = 0
        start = True
        successful = True
        notifications = await self.sunverse_client.fetch_latest_notifications()
        required_keys = [
            "pinned",
            "noticeId",
            "exposed",
            "title",
            "body",
            "noticeType",
            "published",
            "parentId",
            "hideFromArtist",
            "membershipOnly",
            "publishAt",
            "exposedStatus",
            "plainBody",
            "attachment",
            "shareUrl",
            "appUrl",
        ]
        attachment_keys = ["photo", "video"]

        while loop <= 100:
            if not start:
                await asyncio.sleep(0.25)
                data = await self.sunverse_client._fetcher(self.sunverse_client._urls.latest_notifications_url())
                data = await self.sunverse_client._fetcher(self.sunverse_client._urls.notification_url(int(data["paging"]["nextParams"]["after"])))
                notifications = [await self.sunverse_client._object_creator(ObjectTypes.NOTIFICATION, d, self.sunverse_client._urls.notification_url(int(data["paging"]["nextParams"]["after"]))) for d in data["data"] if d.get("community")]

            for notification in notifications:
                await asyncio.sleep(0.25)
                if notification.post_type != PostTypes.NOTICE:
                    continue

                notice = await self.sunverse_client.fetch_notice(notification.post_id)

                for required_key in required_keys:
                    if required_key not in notice.data:
                        raise KeyError(
                            f"Required Key: {required_key} does not exist "
                            f"in the fetched data for {notice.url}"
                        )

                for key in notice.data:
                    if key not in required_keys:
                        raise KeyError(
                            f"Undefined Key: {key} does not exist in the required "
                            f"keys list for {notice.url} - which means that it's a "
                            "new key added on Weverse's side and "
                            "an update to the Notice class is required."
                        )

                for attachment_key in notice.data["attachment"]:
                    if attachment_key not in attachment_keys:
                        raise KeyError(
                            f"Undefined Attachment Key: {attachment_key} does not "
                            f"exist in the attachment keys list for {notice.url}"
                        )

        start = False
        loop += 1

        assert successful

"""async def test(self):
        while True:
            if not start:
                if post.data["membershipOnly"] == True:
                    print(f"Key: membershipOnly is True in {post.url}")

                if post.data["hideFromArtist"] == True:
                    print(f"Key: hideFromArtist is True in {post.url}")

                if post.data["hasProduct"] == True:
                    print(f"Key: hasProduct is True in {post.url}")

                if post.data["locked"] == True:
                    print(f"Key: locked is True in {post.url}")

                if post.data["sectionType"] != "ARTIST":
                    print(f"Key: sectionType is not ARTIST in {post.url}")

                if post.data["postType"] != "NORMAL":
                    print(f"Key: postType is not NORMAL in {post.url}")

                if post.data["tags"]:
                    print(f"Key: tags is not empty in {post.url}")

                if post.data["extension"]:
                    print(f"Key: extension is not empty in {post.url}")

            start = False
            loop += 1

    async def test_two(self):
        start = True
        loop = 0
        data = (await self._fetcher(self._urls.latest_fan_posts_url(47)))["feedTabPosts"]
        required_keys = [
            "extension",
            "shareUrl",
            "bookmarked",
            "locked",
            "hasProduct",
            "attachment",
            "body",
            "tags",
            "availableActions",
            "postType",
            "sectionType",
            "publishedAt",
            "hideFromArtist",
            "membershipOnly",
            "commentCount",
            "postId",
            "emotionCount",
            "plainBody",
            "author",
            "community",
        ]
        attachment_keys = ["photo", "video"]
        optional_keys = ["artistReactions", "viewerEmotionId"]

        while True:
            if not start:
                data = await self._fetcher(self._urls.test_url(data["paging"]["nextParams"]["after"]))

            for d in data["data"]:
                await asyncio.sleep(0.5)
                post = await self.fetch_post(d["postId"])

                for required_key in required_keys:
                    if required_key not in post.data:
                        print(f"Required Key: {required_key} does not exist in the fetched data for {post.url}")

                for key in post.data:
                    if key not in required_keys and key not in optional_keys:
                        print(f"Undefined Key: {key} does not exist in the required keys list for {post.url}")

                for attachment_key in post.data["attachment"]:
                    if attachment_key not in attachment_keys:
                        print(f"Undefined Attachment Key: {attachment_key} does not exist in the attachment keys list for {post.url}")

                if post.data["membershipOnly"] == True:
                    print(f"Key: membershipOnly is True in {post.url}")

                if post.data["hasProduct"] == True:
                    print(f"Key: hasProduct is True in {post.url}")

                if post.data["locked"] == True:
                    print(f"Key: locked is True in {post.url}")

                if post.data["postType"] != "NORMAL":
                    print(f"Key: postType is not NORMAL in {post.url}")

                if post.data["extension"]:
                    print(f"Key: extension is not empty in {post.url}")

            start = False
            loop += 1"""


class LegacyTests:
    sunverse_client = SunverseClient("", "")

    @pytest.mark.asyncio
    async def test_fetch_community(self, mocker):
        """This test is to ensure that the :class:`sunverse.objects.Community`
        object is able to be created successfully.

        Notes
        -----
        The data used for testing this method is mocked. As such, no API calls
        will be made when this test executes.
        """
        with open("tests/test_community.json", encoding="utf-8") as file:
            community_data = json.load(file)

        mocker.patch(
            "sunverse.sunverse.SunverseClient.fetch_community",
            return_value=Community(community_data),
        )

        community = await self.sunverse_client.fetch_community(1029)
        assert community == Community(community_data)

    @pytest.mark.asyncio
    async def test_fetch_artists(self, mocker):
        """This test is to ensure that the :class:`sunverse.objects.Artist`
        object is able to be created successfully.

        Notes
        -----
        The data used for testing this method is mocked. As such, no API calls
        will be made when this test executes.
        """

        with open("tests/test_artists.json", encoding="utf-8") as file:
            artist_data = json.load(file)

        mocker.patch(
            "sunverse.sunverse.SunverseClient.fetch_artists",
            return_value={data["memberId"]: Artist(data, 1029) for data in artist_data},
        )

        artists = await self.sunverse_client.fetch_artists(1029)
        assert artists == Artist(artist_data, 1029)

    @pytest.mark.asyncio
    async def test_fetch_notification(self, mocker):
        """This test is to ensure that the :class:`sunverse.objects.Notification`
        object is able to be created successfully.

        Notes
        -----
        The data used for testing this method is mocked. As such, no API calls
        will be made when this test executes.
        """
        with open("tests/test_notification.json", encoding="utf-8") as file:
            notification_data = json.load(file)

        mocker.patch(
            "sunverse.sunverse.SunverseClient.fetch_notification",
            return_value=Notification(notification_data),
        )

        notification = await self.sunverse_client.fetch_notification(1576070310486122793)
        assert notification == Notification(notification_data)



from typing import Optional

from sunverse.objects.attachments import Photo, Snippet, Video


class Postlike:
    """Represents a Weverse Post-Like Object.

    Post-Like Objects refer to the different types of Weverse contents that
    share a similar data structure with each other. Examples of which are posts
    that fall under the categories of `Post`, `Moment`, `Media` and `Live`.

    Inherited by:

    - :class:`sunverse.objects.Post`
    - :class:`sunverse.objects.Medialike`
    - :class:`sunverse.objects.ImageMedia`
    - :class:`sunverse.objects.YoutubeMedia`
    - :class:`sunverse.objects.WeverseMedia`
    - :class:`sunverse.objects.Live`

    .. container:: operations

        .. describe:: x == y

            Checks if two post-like objects are equal.

        .. describe:: x != y

            Checks if two post-like objects are not equal.

        .. describe:: hash(x)

            Returns the post-like object's hash.

    Attributes
    ----------
    data: :class:`dict`
        The raw data directly taken from the response generated by Weverse's API.
    id: :class:`int`
        The ID of the post-like object.
    like_id: Optional[:class:`str`]
        The ID of the like on the post-like object, if the user has liked the
        post-like object.
    body: :class:`str`
        The body that is displayed on the https://weverse.io website.
        Consider using :attribute:`plain_body` if you do not want markdowns
        and unnecessary information.
    plain_body: :class:`str`
        The plain body of the post-like object that does not have markdowns
        and unnecessary information.
    url: :class:`str`
        The URL that leads to the post-like object.
    extension: :class:`dict`
        Any additional information in the post will be stored in this
        attribute. However, most if not all posts of the
        :class:`sunverse.objects.Post` object do not have any data
        stored in this attribute. Only posts of other objects like
        :class:`sunverse.objects.Media` would have data stored in this
        attribute.
    bookmarked: :class:`bool`
        Whether the user has bookmarked the post-like object.
    locked: :class:`bool`
        UNDETERMINED FUNCTIONALITY: (Has always returned `False`.)
    has_product: :class:`bool`
        UNDETERMINED FUNCTIONALITY: (Has always returned `False`.)
    hide_from_artist: :class:`bool`
        Whether the post-like object is hidden from artists. Will most likely
        ever return `True` in posts posted by non-artists.
    membership_only: :class:`bool`
        Whether the post-like object can only be seen by users who have a
        paid membership in the community the post-like object belongs to.
    like_count: :class:`int`
        The number of likes for the post-like object.
    comment_count: :class:`int`
        The number of comments for the post-like object.
    published_at: :class:`int`
        The time the post-like object got created at, in epoch.
    post_type: :class:`str`
        The post type of the post-like object.
    section_type: :class:`str`
        The section the post-type object falls under.
    tags: list[:class:`str`]
        The list of hashtags used in the post-like object.
    community_id: :class:`int`
        The community ID of the community the post-like object belongs to.
    author_id: :class:`int`
        The author ID of the author who wrote the post-like object.
    """

    __slots__ = (
        "data",
        "id",
        "like_id",
        "body",
        "plain_body",
        "url",
        "extension",
        "bookmarked",
        "locked",
        "has_product",
        "hide_from_artist",
        "membership_only",
        "like_count",
        "comment_count",
        "published_at",
        "post_type",
        "section_type",
        "tags",
        "community_id",
        "author_id",
    )

    def __init__(self, data: dict):
        self.data: dict = data
        self.id: str = data["postId"]
        self.like_id: Optional[str] = data.get("viewerEmotionId")
        self.body: str = data["body"]
        self.plain_body: str = data["plainBody"]
        self.url: str = data["shareUrl"]
        self.extension: dict = data["extension"]
        self.bookmarked: bool = data["bookmarked"]
        self.locked: bool = data["locked"]
        self.has_product: bool = data["hasProduct"]
        self.hide_from_artist: bool = data["hideFromArtist"]
        self.membership_only: bool = data["membershipOnly"]
        self.like_count: int = data["emotionCount"]
        self.comment_count: int = data["commentCount"]
        self.published_at: int = data["publishedAt"]
        self.post_type: str = data["postType"]
        self.section_type: str = data["sectionType"]
        self.tags: list[str] = data["tags"]
        self.community_id: int = data["community"]["communityId"]
        self.author_id: str = data["author"]["memberId"]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __hash__(self):
        return hash(self.id)


class Post(Postlike):
    """Represents a Weverse Post. Inherits from :class:`sunverse.objects.Postlike`.

    Shares the same attributes with :class:`sunverse.objects.Postlike`.

    .. container:: operations

        .. describe:: str(x)

            Returns the post's plain body.
    """

    def __repr__(self):
        return f"Post post_id={self.id}, plain_body={self.plain_body}"

    def __str__(self):
        return self.plain_body

    @property
    def photos(self) -> list[Photo]:
        """list[:class:`sunverse.objects.Photo`]: A list of
        :class:`sunverse.objects.Photo` objects in the post.

        Returns an empty list if there are no photos.
        """
        if not self.data["attachment"].get("photo"):
            return []

        return [
            Photo(photo_data)
            for photo_data in self.data["attachment"]["photo"].values()
        ]

    @property
    def videos(self) -> list[Video]:
        """list[:class:`sunverse.objects.Video`]: A list of
        :class:`sunverse.objects.Video` objects in the post.

        Returns an empty list if there are no videos.
        """
        if not self.data["attachment"].get("video"):
            return []

        return [
            Video(video_data)
            for video_data in self.data["attachment"]["video"].values()
        ]

    @property
    def snippets(self) -> list[Snippet]:
        """list[:class:`sunverse.objects.Snippet`]: A list of
        :class:`sunverse.objects.Snippet` objects in the post.

        Returns an empty list if there are no photos.
        """
        if not self.data["attachment"].get("snippet"):
            return []

        return [
            Snippet(snippet_data)
            for snippet_data in self.data["attachment"]["snippet"].values()
        ]



import base64
import hmac
import time
import urllib.parse
from hashlib import sha1

WEVERSE_KEY = "1b9cb6378d959b45714bec49971ade22e6e24e42"


def get_current_epoch_time() -> int:
    """Returns the current epoch time in milli seconds."""
    return int(time.time() * 1000)


def generate_message_digest(message: bytes) -> str:
    """Generates the message digest for Weverse.

    Parameters
    ----------
    message: :class:`bytes`
        The message to generate the digest for, in bytes.

    Returns
    -------
    :class:`str`
        The message digest made URL-compatible.
    """
    key = bytes(WEVERSE_KEY, "utf-8")
    hashed = hmac.new(key=key, msg=message, digestmod=sha1)
    message_digest = base64.b64encode(hashed.digest())
    return urllib.parse.quote_plus(message_digest)


class URLHandler:
    def __init__(self):
        self.login_url = (
            "https://accountapi.weverse.io/web/api/v2/auth/token/by-credentials"
        )
        self._base_url = "https://global.apis.naver.com/weverse/wevweb"
        self._parameters = (
            "?appId=be4d79eb8fc7bd008ee82c8ec4ff6fd4&language=en&platform=WEB&wpf=pc"
        )

    def _form_request_url(self, url: str) -> str:
        indexed_url = url[0:255]
        epoch_time = get_current_epoch_time()

        url_with_epoch = f"{indexed_url}{epoch_time}".encode("utf-8")

        message_digest = generate_message_digest(url_with_epoch)
        return f"{self._base_url}{url}&wmsgpad={epoch_time}&wmd={message_digest}"

    def joined_communities_url(self) -> str:
        """The endpoint for fetching the data from Weverse
        that contains the communities the user have joined.
        """
        url = "/noti/feed/v1.0/activities/community" + self._parameters
        return self._form_request_url(url)

    def community_url(self, community_id: int) -> str:
        """The endpoint for fetching the data of a specified Community from Weverse."""
        url = (
            f"/community/v1.0/community-{community_id}"
            + self._parameters
            + "&fieldSet=communityHomeV1"
        )
        return self._form_request_url(url)

    def artists_url(self, community_id: int) -> str:
        """The endpoint for fetching the data of the artists of a specified Community
        from Weverse."""
        url = (
            f"/member/v1.0/community-{community_id}/artistMembers"
            + self._parameters
            + "&fieldSet=artistMembersV1&fields=communityId"
            "%2CjoinedDate%2CprofileName%2CprofileImageUrl"
            "%2CprofileCoverImageUrl%2CprofileComment"
        )
        return self._form_request_url(url)

    def latest_notifications_url(self) -> str:
        """The endpoint for fetching the data of the latest notifications
        from Weverse."""
        url = "/noti/feed/v1.0/activities" + self._parameters
        return self._form_request_url(url)

    def notification_url(self, notification_id: int) -> str:
        """The endpoint for fetching the data of a specified Notification
        from Weverse."""
        url = (
            "/noti/feed/v1.0/activities"
            + self._parameters
            + f"&next={notification_id+1}"
        )
        return self._form_request_url(url)

    def post_url(self, post_id: str) -> str:
        """The endpoint for fetching the data of a specified Post from Weverse."""
        url = f"/post/v1.0/post-{post_id}" + self._parameters + "&fieldSet=postV1"
        return self._form_request_url(url)

    def latest_fan_posts_url(self, community_id: int) -> str:
        """The endpoint for fetching the data of the latest fan posts
        from Weverse."""
        url = (
            f"/post/v1.0/community-{community_id}/feedTab"
            + self._parameters
            + "&fields=feedTabPosts.fieldSet%28postsV1%29.limit%2820%29%2CcontentSlots."
            "fieldSet%28contentSlotV1%29&platform=WEB"
        )
        return self._form_request_url(url)

    def video_url(self, video_id: str) -> str:
        """The endpoint for fetching the URL of videos from Weverse."""
        url = f"/cvideo/v1.0/cvideo-{video_id}/downloadInfo" + self._parameters
        return self._form_request_url(url)

    def notice_url(self, notice_id: int) -> str:
        """The endpoint for fetching the data of a specified Notice from Weverse."""
        url = (
            f"/notice/v1.0/notice-{notice_id}" + self._parameters + "&fieldSet=noticeV1"
        )
        return self._form_request_url(url)

    def member_url(self, member_id: str) -> str:
        """The endpoint for fetching the data of a specified Member from Weverse."""
        url = (
            f"/member/v1.0/member-{member_id}"
            + self._parameters
            + "&fields=memberId%2CcommunityId%2Cjoined%2CjoinedDate%2CprofileType"
            "%2CprofileName%2CprofileImageUrl%2CprofileCoverImageUrl%2CprofileComment"
            "%2Chidden%2Cblinded%2CmemberJoinStatus%2CfollowCount"
            "%2ChasMembership%2ChasOfficialMark"
        )
        return self._form_request_url(url)

    def artist_comments_url(self, post_id: str) -> str:
        """The endpoint for fetching the data of the artist comments on a
        specified post from Weverse."""
        url = (
            f"/comment/v1.0/post-{post_id}/artistComments"
            + self._parameters
            + "&fieldSet=postArtistCommentsV1"
        )
        return self._form_request_url(url)

    def comment_url(self, comment_id: str) -> str:
        """The endpoint for fetching the data of a specified Comment from Weverse."""
        url = (
            f"/comment/v1.0/comment-{comment_id}"
            + self._parameters
            + "&fieldSet=commentV1"
        )
        return self._form_request_url(url)



class Community:
    """Represents a Weverse Community.

    .. container:: operations

        .. describe:: x == y

            Checks if two communities are equal.

        .. describe:: x != y

            Checks if two communities are not equal.

        .. describe:: hash(x)

            Returns the community's hash.

        .. describe:: str(x)

            Returns the community's name.

    Attributes
    ----------
    data: :class:`dict`
        The raw data directly taken from the response generated by Weverse's API.
    id: :class:`int`
        The ID of the community. This can be used to fetch the dictionary
        that contains the :class:`objects.artist.Artist` objects using
        the :class:`SunverseClient.fetch_artists()` method.
    name: :class:`str`
        The official name of the community.
    alias: :class:`str`
        The alias of the community.
    code: :class:`str` | :class:`None`
        The code of the community, if any.
    url_path: :class:`str`
        The URL path to the community. E.g. lesserafim which can be used to
        form a proper URL.
    profile_image_url: :class:`str`
        The URL of the profile image of the community.
    profile_name: :class:`str`
        The name shown on the profile of the community.
    profile_cover_image_url: :class:`str`
        The URL of the profile cover image of the community.
    logo_image_url: :class:`str`
        The URL to the logo image of the community.
    home_header_image_url: :class:`str`
        The URL to the home header image of the community.
    fandom_name: :class:`str` | :class:`None`
        The name of the fandom of the community, if any.
    fan_event_url: :class:`str` | :class:`None`
        The URL used to participate in the fan events of the community, if any.
    home_gradation_colour: :class:`dict`
        The colour gradation used for the homepage for the community.
    has_membership_product: :class:`bool`
        Whether the community sells any paid membership product.
    member_count: :class:`int` | :class:`None`
        The number of users who have joined the community, if not hidden.
    birthday_artists: list[:class:`str`](?)
        The list of birthday artists of the community.
    """

    __slots__ = (
        "data",
        "id",
        "name",
        "alias",
        "code",
        "url_path",
        "profile_image_url",
        "profile_name",
        "profile_cover_image_url",
        "logo_image_url",
        "home_header_image_url",
        "fandom_name",
        "fan_event_url",
        "home_gradation_colour",
        "has_membership_product",
        "member_count",
        "birthday_artists",
    )

    def __init__(self, data: dict):
        self.data: dict = data
        self.id: int = data["communityId"]
        self.name: str = data["communityName"]
        self.alias: str = data["communityAlias"]
        self.code: str | None = data.get("artistCode")
        self.url_path: str = data["urlPath"]
        self.profile_image_url: str = data["agencyProfile"]["profileImageUrl"]
        self.profile_name: str = data["agencyProfile"]["profileName"]
        self.profile_cover_image_url: str = data["agencyProfile"][
            "profileCoverImageUrl"
        ]
        self.logo_image_url: str = data["logoImage"]
        self.home_header_image_url: str = data["homeHeaderImage"]
        self.fandom_name: str | None = data.get("fandomName")
        self.fan_event_url: str | None = data.get("fanEventUrl")
        self.home_gradation_colour: dict = data["homeGradationColor"]
        self.has_membership_product: bool = data["hasMembershipProduct"]
        self.member_count: int | None = data.get("memberCount")
        self.birthday_artists: list = data["birthdayArtists"]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __repr__(self):
        return f"Community community_id={self.id}, name={self.name}"

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.id)



class Photo:
    """Represents a Weverse Photo.

    .. container:: operations

        .. describe:: x == y

            Checks if two photos are equal.

        .. describe:: x != y

            Checks if two photos are not equal.

        .. describe:: hash(x)

            Returns the photo's hash.

        .. describe:: str(x)

            Returns the photo's URL.

    Attributes
    ----------
    data: :class:`dict`
        The raw data directly taken from the response generated by Weverse's API.
    id: :class:`int`
        The ID of the photo.
    url: :class:`str`
        The URL of the photo.
    width: :class:`int`
        The width of the photo.
    height: :class:`int`
        The height of the photo.
    """

    __slots__ = (
        "data",
        "id",
        "url",
        "width",
        "height",
    )

    def __init__(self, data: dict):
        self.data: dict = data
        self.id: str = data["photoId"]
        self.url: str = data["url"]
        self.width: int = data["width"]
        self.height: int = data["height"]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __repr__(self):
        return f"Photo photo_id={self.id}, url={self.url}"

    def __str__(self):
        return self.url

    def __hash__(self):
        return hash(self.id)


class Video:
    """Represents a Weverse Video.

    .. container:: operations

        .. describe:: x == y

            Checks if two videos are equal.

        .. describe:: x != y

            Checks if two videos are not equal.

        .. describe:: hash(x)

            Returns the video's hash.

        .. describe:: str(x)

            Returns the video's duration.

    Attributes
    ----------
    data: :class:`dict`
        The raw data directly taken from the response generated by Weverse's API.
    id: :class:`int`
        The ID of the video.
    width: :class:`int`
        The width of the video.
    height: :class:`int`
        The height of the video.
    duration: :class:`int`
        The duration of the video, in seconds.
    thumbnail_url: :class:`str`
        The URL of the thumbnail of the video.

    Notes
    -----
    Due to Weverse's implementations, the URL of the video itself is not accessible
    from this object. As a workaround, please make use of the
    :class:`SunverseClient.fetch_video_url()` method to make an additional API
    call to fetch the video URL.
    """

    __slots__ = (
        "data",
        "id",
        "width",
        "height",
        "duration",
        "thumbnail_url",
    )

    def __init__(self, data: dict):
        self.data: dict = data
        self.id: str = data["videoId"]
        self.width: int = data["uploadInfo"]["width"]
        self.height: int = data["uploadInfo"]["height"]
        self.duration: int = data["uploadInfo"]["playTime"]
        self.thumbnail_url: str = data["uploadInfo"]["imageUrl"]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __repr__(self):
        return f"Video video_id={self.id} duration={self.duration}"

    def __str__(self):
        return str(self.duration)

    def __hash__(self):
        return hash(self.id)


class Snippet:
    """Represents a Weverse Snippet. A Snippet can be
    perceived as an embed.

    .. container:: operations

        .. describe:: x == y

            Checks if two snippets are equal.

        .. describe:: x != y

            Checks if two snippets are not equal.

        .. describe:: hash(x)

            Returns the snippet's hash.

        .. describe:: str(x)

            Returns the snippet's title.

    Attributes
    ----------
    data: :class:`dict`
        The raw data directly taken from the response generated by Weverse's API.
    id: :class:`int`
        The ID of the snippet.
    url: :class:`str`
        The URL of the website the snippet brings to.
    title: :class:`str`
        The title of the snippet.
    description: :class:`str`
        The description of the snippet.
    type: :class:`str` | :class:`None`
        The type of snippet it is, if any.
    site: :class:`str` | :class:`None`
        The website of the snippet, if any.
    domain: :class:`str`
        The domain name of the snippet.
    thumbnail_url: :class:`str` | :class:`None`
        The URL of the thumbnail of the snippet, if any.
    """

    __slots__ = (
        "data",
        "id",
        "url",
        "title",
        "description",
        "type",
        "site",
        "domain",
        "thumbnail_url",
    )

    def __init__(self, data: dict):
        self.data: dict = data
        self.id: str = data["snippetId"]
        self.url: str = data["url"]
        self.title: str = data["title"]
        self.description: str = data["description"]
        self.type: str | None = data.get("type")
        self.site: str | None = data.get("site")
        self.domain: str = data["domain"]
        self.thumbnail_url: str | None = (
            data["image"]["url"] if data.get("image") else None
        )

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __repr__(self):
        return f"Snippet snippet_id={self.id} title={self.title}"

    def __str__(self):
        return self.title

    def __hash__(self):
        return hash(self.id)



from enum import Enum


class PostTypes(Enum):
    POST = "post"
    ARTIST_POST_COMMENT = "artist_post_comment"
    USER_POST_COMMENT = "user_post_comment"
    MEDIA_COMMENT = "media_comment"
    MOMENT_COMMENT = "moment_comment"
    MOMENT = "moment"
    LIVE = "live"
    NOTICE = "notice"
    MEDIA = "media"
    BIRTHDAY = "birthday"

    def __repr__(self):
        return self.value


class ObjectTypes(Enum):
    COMMUNITY = "community"
    ARTIST = "artist"
    NOTIFICATION = "notification"
    POST = "post"
    MEDIA = "media"
    IMAGE_MEDIA = "image_media"
    WEVERSE_MEDIA = "weverse_media"
    YOUTUBE_MEDIA = "youtube_media"
    LIVE = "live"
    MOMENT = "moment"
    NEW_MOMENT = "new_moment"
    OLD_MOMENT = "old_moment"
    NOTICE = "notice"
    MEMBER = "member"
    COMMENT = "comment"

    def __repr__(self):
        return self.value



class BaseParent:
    """Represents a Weverse Base Parent class.

    Inherited by:

    - :class:`objects.comment.ParentComment`
    - :class:`objects.comment.ParentPost`
    - :class:`objects.comment.ParentMediaPost`

    Attributes
    ----------
    data: :class:`str`
        The raw data directly taken from the response generated by Weverse's API.
    body: :class:`str`
        The body of the base parent object.
    author: :class:`objects.comment.BaseAuthor`
        The :class:`objects.comment.BaseAuthor` object of the parent author
        who wrote the base parent object.
    """

    __slots__ = ("data", "body", "author")

    def __init__(self, data: dict):
        self.data: dict = data
        self.body: str = data["body"]
        self.author: BaseAuthor = BaseAuthor(data["author"])


class ParentComment(BaseParent):
    """Represents a Weverse Parent Comment. Inherits from
    :class:`objects.comment.BaseParent`.

    Shares the same attributes with :class:`objects.comment.BaseParent`.

    .. container:: operations

        .. describe:: x == y

            Checks if two parent comments are equal.

        .. describe:: x != y

            Checks if two parent comments are not equal.

        .. describe:: hash(x)

            Returns the parent comment's hash.

        .. describe:: str(x)

            Returns the parent comment's body.

    Attributes
    ----------
    id: :class:`str`
        The ID of the parent comment. If more information about the parent
        comment is required by the user, the :class:`SunverseClient.fetch_comment()`
        method can be used to fetch the actual :class:`objects.comment.Comment`
        object.
    created_at: :class:`int`
        The time the parent comment got created at, in epoch.
    """

    __slots__ = ("id", "created_at")

    def __init__(self, data: dict):
        super().__init__(data)
        self.id: str = data["commentId"]
        self.created_at: int = data["createdAt"]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __repr__(self):
        return f"Parent Comment comment_id={self.id}, body={self.body}"

    def __str__(self):
        return self.body

    def __hash__(self):
        return hash(self.id)


class ParentPost(BaseParent):
    """Represents a Weverse Parent Post. Inherits from
    :class:`objects.comment.BaseParent`.

    Shares the same attributes with :class:`objects.comment.BaseParent`.

    Inherited by:

    - :class:`objects.comment.ParentMediaPost`
    - :class:`objects.comment.RootPost`

    .. container:: operations

        .. describe:: x == y

            Checks if two parent posts are equal.

        .. describe:: x != y

            Checks if two parent posts are not equal.

        .. describe:: hash(x)

            Returns the parent post's hash.

        .. describe:: str(x)

            Returns the parent post's plain body.

    Attributes
    ----------
    id: :class:`str`
        The ID of the parent post. If more information about the parent
        post is required by the user, the :class:`Sunverse.Client.fetch_post()`
        method can be used to fetch the actual :class:`objects.post.Post` object.
    plain_body: :class:`str`
        The plain body of the parent post that does not have markdowns
        and unnecessary information.
    post_type: :class:`str`
        The post type of the parent post.
    """

    __slots__ = ("id", "plain_body", "post_type")

    def __init__(self, data: dict):
        super().__init__(data)
        self.id: str = data["postId"]
        self.plain_body: str = data["plainBody"]
        self.post_type: str = data["postType"]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __repr__(self):
        return f"Parent Post post_id={self.id}, body={self.plain_body}"

    def __str__(self):
        return self.plain_body

    def __hash__(self):
        return hash(self.id)


class ParentMediaPost(ParentPost):
    """Represents a Weverse Parent Media Post. Inherits from
    :class:`objects.comment.BaseParent` and :class:`objects.comment.ParentPost`.

    Shares the same attributes with :class:`objects.comment.BaseParent` and
    :class:`objects.comment.ParentPost`.

    .. container:: operations

        .. describe:: str(x)

            Returns the parent media post's title.

    Attributes
    ----------
    title: :class:`str`
        The title of the parent media post.
    """

    __slots__ = ("title",)

    def __init__(self, data: dict):
        super().__init__(data)
        self.title: str = data["title"]

    def __repr__(self):
        return f"Parent Media Post post_id={self.id}, title={self.title}"

    def __str__(self):
        return self.title


class RootPost(ParentPost):
    """Represents a Weverse Root Post. Inherits from
    :class:`objects.comment.BaseParent` and :class:`objects.comment.ParentPost`.

    Shares the same attributes with :class:`objects.comment.BaseParent` and
    :class:`objects.comment.ParentPost`.

    Attributes
    ----------
    url: :class:`str`
        The URL that leads to the root post.
    section_type: :class:`str`
        The section the root post falls under.
    author: :class:`objects.comment.Author`
        The :class:`objects.comment.Author` object of the author
        who wrote the root post.
    """

    __slots__ = ("url", "section_type")

    def __init__(self, data: dict):
        super().__init__(data)
        self.url: str = data["shareUrl"]
        self.section_type: str = data["sectionType"]
        self.author: Author = Author(data["author"])

    def __repr__(self):
        return f"Root Post post_id={self.id}, body={self.plain_body}"


class RootMediaPost(ParentMediaPost):
    """Represents a Weverse Root Media Post. Inherits from
    :class:`objects.comment.BaseParent`, :class:`objects.comment.ParentPost`
    and :class:`objects.comment.ParentMediaPost`.

    Shares the same attributes with :class:`objects.comment.BaseParent`,
    :class:`objects.comment.ParentPost` and :class:`objects.comment.ParentMediaPost`.

    Attributes
    ----------
    url: :class:`str`
        The URL that leads to the root media post.
    section_type: :class:`str`
        The section the root media post falls under.
    author: :class:`objects.comment.Author`
        The :class:`objects.comment.Author` object of the author
        who wrote the root media post.
    """

    __slots__ = ("url", "section_type")

    def __init__(self, data: dict):
        super().__init__(data)
        self.url: str = data["shareUrl"]
        self.section_type: str = data["sectionType"]
        self.author: Author = Author(data["author"])

    def __repr__(self):
        return f"Root Media Post post_id={self.id}, title={self.title}"


class Comment:
    """Represents a Weverse Comment.

    .. container:: operations

        .. describe:: x == y

            Checks if two comments are equal.

        .. describe:: x != y

            Checks if two comments are not equal.

        .. describe:: hash(x)

            Returns the comment's hash.

        .. describe:: str(x)

            Returns the comment's body.

    Attributes
    ----------
    data: :class:`str`
        The raw data directly taken from the response generated by Weverse's API.
    id: :class:`str`
        The ID of the comment.
    body: :class:`str`
        The body of the comment.
    comment_count: :class:`int`
        The number of comments for the comment.
    like_count: :class:`int`
        The number of likes for the comment.
    created_at: :class:`int`
        The time the comment got created at, in epoch.
    author: :class:`objects.comment.Author`
        The :class:`objects.comment.Author` object of the author
        who wrote the comment.
    """

    __slots__ = (
        "data",
        "id",
        "body",
        "comment_count",
        "like_count",
        "created_at",
        "author",
    )

    def __init__(self, data: dict):
        self.data: dict = data
        self.id: str = data["commentId"]
        self.body: str = data["body"]
        self.comment_count: int = data["commentCount"]
        self.like_count: int = data["emotionCount"]
        self.created_at: int = data["createdAt"]
        self.author: Author = Author(data["author"])

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __repr__(self):
        return f"Comment comment_id={self.id}, body={self.body}"

    def __str__(self):
        return self.body

    def __hash__(self):
        return hash(self.id)

    @property
    def parent(self) -> ParentComment | ParentMediaPost | ParentPost:
        """:class:`objects.comment.ParentComment` |
        :class:`objects.comment.ParentPost` |
        :class:`objects.comment.ParentMediaPost`:

        - Returns the :class:`objects.comment.ParentComment` object if the\
        parent is a comment.

        - Returns the :class:`objects.comment.ParentMediaPost` object if the\
        parent is a media post.

        - Returns the :class:`objects.comment.ParentPost` object if the parent\
        is a post.
        """
        post_types = ["NORMAL", "MOMENT", "OFFICIAL"]

        if self.data["parent"]["type"] == "COMMENT":
            return ParentComment(self.data["parent"]["data"])

        if self.data["parent"]["type"] == "POST":
            if self.data["parent"]["data"]["postType"] in post_types:
                return ParentPost(self.data["parent"]["data"])

            else:
                return ParentMediaPost(self.data["parent"]["data"])

    @property
    def root(self) -> RootMediaPost | RootPost:
        """:class:`objects.comment.RootMediaPost` |
        :class:`objects.comment.RootPost`:

        - Returns the :class:`objects.comment.RootMediaPost` object if the\
        root post is a media post.

        - Returns the :class:`objects.comment.RootPost` object if the root\
        post is a normal post.
        """
        post_types = ["NORMAL", "MOMENT", "OFFICIAL"]

        if self.data["root"]["data"]["postType"] in post_types:
            return RootPost(self.data["root"]["data"])

        else:
            return RootMediaPost(self.data["root"]["data"])

    @property
    def url(self) -> str:
        """:class:`str`: Returns the URL to the comment."""
        return self.root.url + f"/comment/{self.id}"


class BaseAuthor:
    """Represents a Weverse Base Author class.

    Inherited by:

    - :class:`objects.comment.Author`

    .. container:: operations

        .. describe:: x == y

            Checks if two base authors are equal.

        .. describe:: x != y

            Checks if two base authors are not equal.

        .. describe:: hash(x)

            Returns the base author's hash.

        .. describe:: str(x)

            Returns the base author's name.

    Attributes
    ----------
    data: :class:`str`
        The raw data directly taken from the response generated by Weverse's API.
    id: :class:`str`
        The ID of the base author. If more information about the author is
        required by the user, the :class:`SunverseClient.fetch_member()` method
        can be used to fetch the actual :class:`objects.member.Member` object.
    name: :class:`str`
        The name of the base author.
    image_url: :class:`str`
        The URL of the profile image of the base author. Returns an empty
        string if the base author does not have a profile image.
    profile_type: :class:`str`
        The profile type of the base author.
    """

    __slots__ = (
        "data",
        "id",
        "name",
        "image_url",
        "profile_type",
    )

    def __init__(self, data: dict):
        self.data: dict = data
        self.id: str = data["memberId"]
        self.name: str = data["profileName"]
        self.image_url: str = (
            data["profileImageUrl"] if data.get("profileImageUrl") else ""
        )
        self.profile_type: str = data["profileType"]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __repr__(self):
        return f"Parent Author author_id={self.id}, name={self.name}"

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.id)


class Author(BaseAuthor):
    """Represents a Weverse Author. Inherits from :class:`objects.comment.BaseAuthor`.

    Shares the same attributes with :class:`objects.comment.BaseAuthor`.

    Attributes
    ----------
    has_official_mark: :class:`bool`
        Whether the author is an official Weverse account.
    community_id: :class:`int`
        The community ID of the community the author belongs to. This
        can be used to fetch the actual :class:`objects.community.Community`
        object if needed using the :class:`SunverseClient.fetch_community()`
        method.
    """

    __slots__ = ("has_offical_mark", "community_id")

    def __init__(self, data: dict):
        super().__init__(data)
        self.has_offical_mark: bool = data["hasOfficialMark"]
        self.community_id: int = data["communityId"]

    def __repr__(self):
        return f"Author author_id={self.id}, name={self.name}"



from sunverse.objects.media import WeverseMedia


class Live(WeverseMedia):
    """Represents a Weverse Live Broadcast.
    Inherits from :class:`objects.media.WeverseMedia`.

    Shares the same attributes with :class:`objects.post.Postlike`,
    :class:`objects.media.Medialike` and :class:`objects.media.WeverseMedia`.

    Attributes
    ----------
    message_count: :class:`int`
        The number of messages in the live broadcast.
    """

    __slots__ = ("message_count",)

    def __init__(self, data: dict):
        super().__init__(data)
        self.message_count: int = data["extension"]["mediaInfo"]["chat"]["messageCount"]

    def __repr__(self):
        return f"Live live_id={self.id}, title={self.title}"



import re

from sunverse.objects.attachments import Photo


class Notice:
    """Represents a Weverse Notice.

    .. container:: operations

        .. describe:: x == y

            Checks if two notice objects are equal.

        .. describe:: x != y

            Checks if two notice objects are not equal.

        .. describe:: hash(x)

            Returns the notice's hash.

        .. describe:: str(x)

            Returns the notice's plain body.

    Attributes
    ----------
    data: :class:`dict`
        The raw data directly taken from the response generated by Weverse's API.
    id: :class:`int`
        The ID of the notice.
    title: :class:`str`
        The title of the notice.
    body: :class:`str`
        The body that is displayed on the https://weverse.io website.
        Consider using :attribute:`plain_body` if you do not want markdowns
        and unnecessary information.
    plain_body: :class:`str`
        The plain body of the notice that does not have markdowns
        and unnecessary information.
    url: :class:`str`
        The URL that leads to the notice.
    exposed: :class:`bool`
        Whether the notice is exposed.
    published: :class:`bool`
        Whether the notice is published.
    hide_from_artist: :class:`bool`
        Whether the notice is hidden from artists.
    membership_only: :class:`bool`
        Whether the notice can only be seen by users who have a
        paid membership in the community the notice belongs to.
    pinned: :class:`bool`
        Whether the notice is pinned.
    published_at: :class:`int`
        The time the notice got created at, in epoch.
    notice_type: :class:`str`
        The type of notice it is.
    exposed_status: :class:`str`
        The exposed status of the notice.
    """

    __slots__ = (
        "data",
        "id",
        "title",
        "body",
        "plain_body",
        "url",
        "exposed",
        "published",
        "hide_from_artist",
        "membership_only",
        "pinned",
        "published_at",
        "notice_type",
        "exposed_status",
    )

    def __init__(self, data: dict):
        self.data: dict = data
        self.id: int = data["noticeId"]
        self.title: str = data["title"]
        self.body: str = data["body"]
        self.plain_body: str = data["plainBody"]
        self.url: str = data["shareUrl"]
        self.exposed: bool = data["exposed"]
        self.published: bool = data["published"]
        self.hide_from_artist: bool = data["hideFromArtist"]
        self.membership_only: bool = data["membershipOnly"]
        self.pinned: bool = data["pinned"]
        self.published_at: int = data["publishAt"]
        self.notice_type: str = data["noticeType"]
        self.exposed_status: str = data["exposedStatus"]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return f"Notice notice_id={self.id}, plain_body={self.plain_body}"

    def __str__(self):
        return self.plain_body

    @property
    def photos(self) -> list[Photo]:
        """list[:class:`sunverse.objects.Photo`]: A list of
        :class:`sunverse.objects.Photo` objects in the notice.

        Returns an empty list if there are no photos.
        """
        if not self.data["attachment"].get("photo"):
            return []

        return [
            Photo(photo_data)
            for photo_data in self.data["attachment"]["photo"].values()
        ]

    @property
    def community_id(self) -> int:
        """:class:`int`: The community ID of the community the notice
        belongs to.
        """
        pattern = re.compile(r"([^a-z -][\d]*)")
        match = re.search(pattern, self.data["parentId"])
        return int(match.group(0))



import re

from setuptools import find_packages, setup


with open("README.md", "r") as f:
    long_description = f.read()

with open("sunverse/__init__.py") as f:
    version = (
        re.search(
            r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE
        ).group(1)
        or ""
    )


def parse_requirements(filename):
    """Load requirements from a requirement file."""
    line_iter = (line.strip() for line in open(filename))
    return [line for line in line_iter if line and not line.startswith("#")]


setup(
    name="Sun Weverse",
    author="Anson Quek",
    url="https://github.com/Anson-Quek/Sun-Weverse",
    version=version,
    packages=find_packages(),
    license="MIT",
    description="Sun Weverse seeks to provide developers with a tool that "
    "allows them to make a bot that is able to retrieve Weverse Posts in "
    "semi real-time with relative ease.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=parse_requirements("requirements.txt"),
    python_requires=">=3.10",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.10",
    ],
)



from typing import Optional
from sunverse.objects.attachments import Photo, Video
from sunverse.objects.post import Postlike


class Momentlike(Postlike):
    """Represents a Weverse Moment-Like Object.

    Inherits from :class:`sunverse.objects.Postlike`.

    Shares the same attributes with :class:`sunverse.objects.Postlike`.

    Inherited by:

    - :class:`sunverse.objects.Moment`
    - :class:`sunverse.objects.OldMoment`

    .. container:: operations

        .. describe:: str(x)

            Returns the moment's plain body.

    Attributes
    ----------
    expire_at: :class:`int`
        The time the moment expires at, in epoch.
    """

    __slots__ = ("expire_at",)

    def __init__(self, data: dict):
        super().__init__(data)

        if "moment" in data["extension"]:
            self.expire_at: int = data["extension"]["moment"]["expireAt"]

        else:
            self.expire_at: int = data["extension"]["momentW1"]["expireAt"]

    def __repr__(self):
        return f"Moment moment_id={self.id}, plain_body={self.plain_body}"

    def __str__(self):
        return self.plain_body


class Moment(Momentlike):
    """Represents a Weverse Moment that has been created after their rework.

    Inherits from :class:`sunverse.objects.Momentlike`.

    Shares the same attributes with :class:`sunverse.objects.Postlike` and
    :class:`sunverse.objects.Moment`.

    Attributes
    ----------
    video: :class:`sunverse.objects.Video`
        The :class:`sunverse.objects.Video` object of the video in the moment.
    """

    __slots__ = ("video",)

    def __init__(self, data: dict):
        super().__init__(data)
        self.video: Video = Video(data["extension"]["moment"]["video"])


class OldMoment(Momentlike):
    """Represents a Weverse Moment that has been created before their rework.

    Inherits from :class:`sunverse.objects.Momentlike`.

    Shares the same attributes with :class:`sunverse.objects.Postlike` and
    :class:`sunverse.objects.Moment`.

    Attributes
    ----------
    photo: Optional[:class:`sunverse.objects.Photo`]
        The :class:`sunverse.objects.Photo` object of the photo in the moment,
        if the image used in the moment is not a default Weverse background image.
    background_image_url: Optional[:class:`str`]
        The URL of the default Weverse background image if it is used.
    """

    __slots__ = ("photo", "background_image_url")

    def __init__(self, data: dict):
        super().__init__(data)
        self.photo: Optional[Photo] = (
            Photo(data["extension"]["momentW1"]["photo"])
            if "photo" in data["extension"]["momentW1"]
            else None
        )
        self.background_image_url: Optional[str] = (
            data["extension"]["momentW1"]["backgroundImageUrl"]
            if "backgroundImageUrl" in data["extension"]["momentW1"]
            else None
        )



class Artist:
    """Represents a Weverse Artist.

    .. container:: operations

        .. describe:: x == y

            Checks if two artists are equal.

        .. describe:: x != y

            Checks if two artists are not equal.

        .. describe:: hash(x)

            Returns the artist's hash.

        .. describe:: str(x)

            Returns the artist's name.

    Attributes
    ----------
    data: :class:`str`
        The raw data directly taken from the response generated by Weverse's API.
    id: :class:`str`
        The ID of the artist.
    image_url: :class:`str`
        The URL of the official image of the artist.
    name: :class:`str`
        The name of the artist.
    birthday: :class:`int`
        The artist's birthday, in epoch time.
    is_birthday: :class:`bool`
        Whether it's the artist's birthday.
    joined_date: :class:`int`
        The date the artist joined the community on, in epoch.
    nickname: :class:`str`
        The nickname of the artist.
    profile_image_url: :class:`str`
        The URL of the profile image of the artist.
    profile_banner_url: :class:`str`
        The URL of the profile banner of the artist.
    profile_comment: :class:`str`
        The profile comment of the artist.
    community_id: :class:`int`
        The community ID of the community the artist belongs to. This
        can be used to fetch the actual :class:`objects.community.Community`
        object if needed using the :class:`SunverseClient.fetch_community()`
        method.
    """

    __slots__ = (
        "data",
        "id",
        "image_url",
        "name",
        "birthday",
        "is_birthday",
        "joined_date",
        "nickname",
        "profile_image_url",
        "profile_banner_url",
        "profile_comment",
        "community_id",
    )

    def __init__(self, data: dict):
        self.data: dict = data
        self.id: str = data["memberId"]
        self.image_url: str = data["artistOfficialProfile"]["officialImageUrl"]
        self.name: str = data["artistOfficialProfile"]["officialName"]
        self.birthday: int = data["artistOfficialProfile"]["birthday"]["date"]
        self.is_birthday: bool = data["artistOfficialProfile"]["birthday"]["bday"]
        self.joined_date: int = data["joinedDate"]
        self.nickname: str = data["profileName"]
        self.profile_image_url: str = data["profileImageUrl"]
        self.profile_banner_url: str = data["profileCoverImageUrl"]
        self.profile_comment: str = data["profileComment"]
        self.community_id: int = data["communityId"]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __repr__(self):
        return f"Artist artist_id={self.id}, name={self.name}"

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.id)



import asyncio
import logging

from aiohttp import ClientConnectionError, ClientResponse, ClientSession
from yarl import URL

from sunverse.objects.notification import Notification

from .enums import ObjectTypes, PostTypes
from .errors import InternalServerError, LoginError, NotFound, RequestFailed
from .objects.artist import Artist
from .objects.comment import Comment
from .objects.community import Community
from .objects.live import Live
from .objects.media import ImageMedia, WeverseMedia, YoutubeMedia
from .objects.member import Member
from .objects.moment import Moment, OldMoment
from .objects.notice import Notice
from .objects.post import Post
from .utils.urls import URLHandler, get_current_epoch_time
from .utils.utilities import MISSING

_LOGIN_HEADERS = {
    "content-type": "application/json",
    "x-acc-app-secret": "5419526f1c624b38b10787e5c10b2a7a",
    "x-acc-app-version": "abc",
    "x-acc-language": "en",
    "x-acc-service-id": "weverse",
    "x-acc-trace-id": "abc",
}

logger = logging.getLogger(__name__)


class SunverseClient:
    """Represents the client connection to Weverse."""

    def __init__(
        self,
        email: str,
        password: str,
        stream_delay: float = 20,
    ):
        self.email: str = email
        self.password: str = password
        self.stream_delay: float = stream_delay

        self._access_token: str = MISSING
        self._urls: URLHandler = URLHandler()
        self._cache: _SunverseCache = _SunverseCache()

        self.__task: asyncio.Task = MISSING
        self.__fetch_delay: float = 0.25

    @property
    def login_payload(self) -> dict:
        """:class:`dict`: Returns the login payload used to login to Weverse."""
        return {"email": self.email, "password": self.password}

    @property
    def api_headers(self) -> dict:
        """:class:`dict`: Returns the API headers used to communicate with Weverse."""
        return {
            "referer": "https://weverse.io",
            "authorization": f"Bearer {self._access_token}",
        }

    async def _fetch_access_token_by_credentials(self) -> None:
        """Fetches and stores the access token using the credentials
        provided by the user. This access token is essential in
        interacting with the Weverse API when pulling data from it.

        Raises
        ------
        ~sunverse.LoginError
            Unable to login successfully.
        """
        async with ClientSession() as client, client.post(
            self._urls.login_url, headers=_LOGIN_HEADERS, json=self.login_payload
        ) as resp:
            if resp.status != 200:
                raise LoginError(resp.status, (await resp.json()).get("message"))

            data = await resp.json()
            self._access_token = data["accessToken"]
            logger.debug(
                "LOGIN SUCCESSFUL\nData: %s\nAccess Token: %s", data, self._access_token
            )

    async def _fetcher(self, url: str) -> dict:
        """Fetches the data from the specified URL provided by the user.

        Parameters
        ----------
        url: :class:`str`
            The URL to fetch the data from.

        Returns
        -------
        :class:`dict`
            A dictionary that contains the raw data of the response provided by
            the Weverse API.
        """
        try:
            async with ClientSession() as sess, sess.get(
                URL(url, encoded=True), headers=self.api_headers
            ) as resp:
                await self._check_response_status(resp)
                return await resp.json()

        except (asyncio.exceptions.TimeoutError, ClientConnectionError):
            logger.warning(
                "CLIENT CONNECTION ERROR IGNORED SILENTLY. "
                "RETRYING TO RETURN REQUESTED RESOURCE."
            )
            return await self._fetcher(url)

        except InternalServerError:
            logger.warning(
                "WEVERSE API ENCOUNTERED TEMPORARY INTERNAL SERVER ERROR. "
                "RETRYING TO RETURN REQUESTED RESOURCE."
            )
            return await self._fetcher(url)

    async def _check_response_status(self, resp: ClientResponse) -> None:
        """Checks the HTTP status code for the response received from Weverse's API.

        Also automatically updates the access token if it is suspected to be
        expired.

        Raises
        ------
        ~sunverse.RequestFailed
            The request failed to get an "OK" or 200 response from the Weverse API.
        ~sunverse.NotFound
            The request received a "NOT FOUND" or 404 response from the Weverse API.
        """
        if resp.status == 401:  # This usually means that the access token is expired.
            return await self._fetch_access_token_by_credentials()

        if resp.status == 404:
            raise NotFound(resp.url)

        if resp.status == 500:
            raise InternalServerError(resp.url)

        if resp.status != 200:  # Raises this exception if other HTTP codes encountered.
            raise RequestFailed(
                resp.url, resp.status, (await resp.json()).get("message")
            )

    async def _object_creator(self, object_type: str, data: dict, retry_url: str):
        """Responsible for creating all the Sunverse Objects based on the
        specified object type.

        Parameters
        ----------
        object_type: :class:`str`
            The type of Sunverse Object to create.
        data: :class:`dict`
            The raw data used for the Sunverse Object's creation.
        retry_url: :class:`str`
            The URL that will be used when retrying due to potentially
            incomplete data provided by the Weverse API.

        Returns
        -------
        :module:`sunverse.objects`
            The Sunverse Objects that are contained in the
            :module:`sunverse.objects` module.

        Notes
        -----
        In an attempt to ensure the data received from the Weverse API is complete,
        this method will also attempt to retry fetching the data from Weverse up to
        3 times should the Sunverse Object creation fail. This is because the Sunverse
        Objects have been tailored to ensure that the data received from the
        Weverse API is complete. As such, if :exception:`KeyError` is encountered
        for whatever reason, it would mean that the data provided by the Weverse API
        is incomplete, and should be retried in an attempt to receive complete data.
        """
        retry_count = 0
        max_retries = 3
        object_mapping = {
            ObjectTypes.COMMUNITY: Community,
            ObjectTypes.ARTIST: Artist,
            ObjectTypes.NOTIFICATION: Notification,
            ObjectTypes.POST: Post,
            ObjectTypes.IMAGE_MEDIA: ImageMedia,
            ObjectTypes.WEVERSE_MEDIA: WeverseMedia,
            ObjectTypes.YOUTUBE_MEDIA: YoutubeMedia,
            ObjectTypes.LIVE: Live,
            ObjectTypes.NEW_MOMENT: Moment,
            ObjectTypes.OLD_MOMENT: OldMoment,
            ObjectTypes.NOTICE: Notice,
            ObjectTypes.MEMBER: Member,
            ObjectTypes.COMMENT: Comment,
        }

        while retry_count < max_retries:
            try:
                # It's basically doing something like Community(data)
                sunverse_object = object_mapping[object_type](data)
                return sunverse_object

            except KeyError:
                data = await self._fetcher(retry_url)
                retry_count += 1

    def _log(self, names_tuple: tuple[str], data: dict, sunverse_object) -> None:
        """A general logger that logs debug fetching messages."""
        first_name, second_name = names_tuple
        logger.debug(
            "%s FETCHED\nData: %s\n%s: %s",
            first_name,
            data,
            second_name,
            sunverse_object,
        )

    async def fetch_new_notification(self) -> tuple[list[Notification], list[Comment]]:
        """Fetches the list of new Weverse Notifications from all
        communities that the signed-in account has access to.

        Returns
        -------
        :class:`list[sunverse.objects.Notification]`:
            A list of :class:`sunverse.objects.Notification` objects the
            signed-in account has access to.

        tuple[list[:class:`sunverse.objects.Notification`],
        list[:class:`sunverse.objects.Comment`]]
            A tuple that contains a list of new :class:`sunverse.objects.Notification`
            objects and a list of new :class:`sunverse.objects.Comment` objects.

        Notes
        -----
        The reason why :class:`sunverse.objects.Comment` objects are returned for
        comment-related notifications and not :class:`sunverse.objects.Notification`
        objects is because Weverse does not provide an easy way to determine which
        comment-related notifications are new. As such, in the post-processing of
        comment-related notifications, the actual :class:`sunverse.objects.Comment`
        objects have to be fetched so that we know which comments are new.
        """
        old_non_comment_noti = dict(self._cache.notification_cache)
        old_comment_noti = dict(self._cache.comment_noti_cache)
        (
            latest_non_comment_noti,
            latest_comment_noti,
        ) = await self.fetch_latest_notifications()

        new_non_comment_noti = self._get_new_non_comment_noti(
            old_non_comment_noti, latest_non_comment_noti
        )

        new_artist_comments = await self._get_new_comment(
            old_comment_noti, latest_comment_noti
        )

        return new_non_comment_noti, new_artist_comments

    @staticmethod
    def _get_new_non_comment_noti(
        old_cache: dict[int, Notification], notifications: list[Notification]
    ) -> list[Notification]:
        """Returns a list of new non-comment notifications that do not exist in
        the cache.

        Parameters
        ----------
        old_cache: dict[:class:`int`, :class:`sunverse.objects.Notificiation`]
            The old notification cache used to check which notifications are new.
        notifications: list[:class:`sunverse.objects.Notification`]
            The list that contains the latest notifications the user has access to.

        Returns
        -------
        list[:class:`sunverse.objects.Notification`]
            A list of new notifications.
        """
        current_epoch_time = get_current_epoch_time()
        epoch_time_difference = 600000  # Equivalent to 10 minutes.
        return [
            notification
            for notification in notifications
            if notification.id not in old_cache
            and current_epoch_time - notification.time_created <= epoch_time_difference
        ]

    async def _get_new_comment(
        self, old_cache: dict[str, int], notifications: list[Notification]
    ) -> list[Comment]:
        """Returns a list of new artist comments.

        Parameters
        ----------
        old_cache: dict[:class:`str`, :class:`int`]
            The old notification cache used to check which notifications are new.
        notifications: list[:class:`sunverse.objects.Notification`]
            The list that contains the latest notifications the user has access to.

        Returns
        -------
        list[:class:`sunverse.objects.Comment`]
            A list of :class:`sunverse.objects.Comment` objects of the new
            artist comments.
        """
        new_artist_comments = []
        current_epoch_time = get_current_epoch_time()
        epoch_time_difference = 60000  # Equivalent to 1 minute

        for notification in notifications:
            await asyncio.sleep(self.__fetch_delay)

            try:
                artist_comments = await self.fetch_artist_comments(notification.post_id)

            except NotFound:  # This is because the Post has been deleted.
                continue

            if not old_cache.get(notification.post_id + notification.author_id):
                for artist_comment in artist_comments:
                    if (
                        current_epoch_time - artist_comment.created_at
                        <= epoch_time_difference
                    ):
                        new_artist_comments.insert(0, artist_comment)

            else:
                # This is done because there can be different artists that
                # have written on a post. As such, a new list is created
                # to contain only comments written by the author provided in
                # the notification data.
                artist_comments = [
                    artist_comment
                    for artist_comment in artist_comments
                    if artist_comment.author.id == notification.author_id
                ]
                comment_count = old_cache[notification.post_id + notification.author_id]
                new_comment_count = notification.count - comment_count

                for artist_comment in artist_comments[0:new_comment_count]:
                    new_artist_comments.insert(0, artist_comment)

        return new_artist_comments

    async def fetch_latest_notifications(self) -> tuple[list, list]:
        """Fetches a tuple that contains a list of the latest non-comment
        notifications and a list of the latest comment notifications from all
        communities that the signed-in account has access to.

        Returns
        -------
        tuple[:class:`list`, :class:`list`]
            A tuple that contains a list of :class:`sunverse.objects.Notification`
            non-comment notification objects and a list of
            :class:`sunverse.objects.Notification` comment notifications objects
            the signed-in account has access to.
        """
        url = self._urls.latest_notifications_url()
        data = await self._fetcher(url)
        (
            non_comment_notifications,
            comment_notifications,
        ) = await self._sort_notifications(data, url)
        self._log(
            ("LATEST NON-COMMENT NOTIFICATIONS", "Non-Comment Notifications"),
            data,
            non_comment_notifications,
        )
        self._log(
            ("LATEST COMMENT NOTIFICATIONS", "Comment Notifications"),
            data,
            comment_notifications,
        )
        return non_comment_notifications, comment_notifications

    async def _sort_notifications(
        self, notification_data: dict, url: str
    ) -> tuple[list, list]:
        """Sorts a dictionary containing notification data into
        comment notifications and non-comment notifications.

        Parameters
        ----------
        notification_data: :class:`dict`
            The dictionary that contains the data of the notifications to sort.
        url: :class:`str`
            The retry URL in case of errors.

        Returns
        -------
        tuple[:class:`list`, :class:`list`]
            A tuple that contains a list of non-comment notifications and
            a list of comment notifications.
        """
        comment_notification_types = [
            PostTypes.USER_POST_COMMENT,
            PostTypes.MEDIA_COMMENT,
            PostTypes.MOMENT_COMMENT,
            PostTypes.ARTIST_POST_COMMENT,
        ]
        non_comment_notifications = []
        comment_notifications = []

        for data in notification_data["data"]:
            if not data.get("community"):
                continue

            notification = await self._object_creator(
                ObjectTypes.NOTIFICATION, data, url
            )

            if notification.post_type in comment_notification_types:
                self._cache.cache_comment_noti(notification)
                comment_notifications.append(notification)

            else:
                self._cache.limitless_cache(ObjectTypes.NOTIFICATION, notification)
                non_comment_notifications.append(notification)

        return non_comment_notifications, comment_notifications

    async def fetch_notification(self, notification_id: int) -> Notification:
        """Fetches a Weverse Notification by its ID.

        Parameters
        ----------
        notification_id: :class:`int`
            The ID of the notification to fetch.

        Returns
        -------
        :class:`sunverse.objects.Notification`
            The Notification object requested by the user.

        Raises
        ------
        ~sunverse.NotFound
            The request received a "NOT FOUND" or 404 response from the Weverse API.
        """
        url = self._urls.notification_url(notification_id)
        data = (await self._fetcher(url))["data"]

        if not data or not data[0].get("community"):
            raise NotFound(url)

        notification = await self._object_creator(
            ObjectTypes.NOTIFICATION, data[0], url
        )

        self._cache.limitless_cache(ObjectTypes.NOTIFICATION, notification)
        self._log(("NOTIFICATION", "Notification"), data, notification)
        return notification

    async def fetch_joined_communities(self) -> list[Community]:
        """Fetches the list of Weverse Communities that the signed-in account
        has access to.

        Returns
        -------
        :class:`list[sunverse.objects.Community]`
        A list of :class:`sunverse.objects.Community` objects the signed-in account
        has access to.
        """
        communities = []
        data = await self._fetcher(self._urls.joined_communities_url())

        for community_data in data["data"]:
            await asyncio.sleep(self.__fetch_delay)
            community = await self.fetch_community(community_data["communityId"])
            communities.append(community)

        self._log(("JOINED COMMUNITIES", "Communities"), data, communities)
        return communities

    async def fetch_community(self, community_id: int) -> Community:
        """Fetches a Weverse Community by its ID.

        Parameters
        ----------
        community_id: :class:`int`
            The ID of the community to fetch.

        Returns
        -------
        :class:`sunverse.objects.Community`
            The Community object requested by the user.
        """
        url = self._urls.community_url(community_id)
        data = await self._fetcher(url)
        community = await self._object_creator(ObjectTypes.COMMUNITY, data, url)
        self._cache.limitless_cache(ObjectTypes.COMMUNITY, community)
        self._log(("COMMUNITY", "Community"), data, community)
        return community

    async def fetch_artists(self, community_id: int) -> dict[str, Artist]:
        """Fetches the Artists that belong to a Weverse Community by the community ID.

        Parameters
        ----------
        community_id: :class:`int`
            The ID of the community that the artists are part of to fetch.

        Returns
        -------
        dict[:class:`str`, :class:`sunverse.objects.Artist`]
            The dictionary that contains the Artist objects requested
            by the user. The keys of the dictionary are the artist ID of artists.
        """
        data = await self._fetcher(self._urls.artists_url(community_id))
        artists: dict[str, Artist] = {
            artist_data["memberId"]: await self._object_creator(
                ObjectTypes.ARTIST,
                artist_data,
                self._urls.artists_url(community_id),
            )
            for artist_data in data
        }

        self._cache.limitless_cache(ObjectTypes.ARTIST, artists)
        self._log(("ARTISTS", "Artists"), data, artists)
        return artists

    async def fetch_post(self, post_id: str) -> Post:
        """Fetches a Weverse Post by its ID.

        Parameters
        ----------
        post_id: :class:`str`
            The ID of the post to fetch.

        Returns
        -------
        :class:`sunverse.objects.Post`
            The Post object requested by the user.
        """
        url = self._urls.post_url(post_id)
        data = await self._fetcher(url)
        post = await self._object_creator(ObjectTypes.POST, data, url)
        self._log(("POST", "Post"), data, post)
        return post

    async def fetch_video_url(self, video_id: str) -> str:
        """Fetches a Weverse Post Video URL by its ID.

        Parameters
        ----------
        video_id: :class:`str`
            The ID of the video to fetch.

        Returns
        -------
        :class:`str`
            The URL of the video.

        Notes
        -----
        An additional API call has to be made because unlike images,
        Weverse does include the URL to videos in the response for posts.
        """
        data = await self._fetcher(self._urls.video_url(video_id))
        url_mapping = {}

        for video_data in data["downloadInfo"]:
            url_mapping[int(video_data["resolution"].replace("P", ""))] = video_data[
                "url"
            ]

        return url_mapping[max(url_mapping)]

    async def fetch_media(
        self, media_id: str
    ) -> ImageMedia | WeverseMedia | YoutubeMedia:
        """Fetches a Weverse Media by its ID.

        Parameters
        ----------
        media_id: :class:`str`
            The ID of the media to fetch.

        Returns
        -------
        Union[:class:`sunverse.objects.ImageMedia`,
        :class:`sunverse.objects.WeverseMedia`, :class:`sunverse.objects.YoutubeMedia`]
            The Media object requested by the user.
        """
        url = self._urls.post_url(media_id)

        try:  # This catches error 403 because the media is paid content.
            data = await self._fetcher(url)

        except RequestFailed:
            logger.warning(
                "FETCHING MEDIA FAILED\nCode: 403\n"
                "Reason: The requested resource is a paid content."
            )
            return

        if "image" in data["extension"]:
            media = await self._object_creator(ObjectTypes.IMAGE_MEDIA, data, url)

        elif "youtube" in data["extension"]:
            media = await self._object_creator(ObjectTypes.YOUTUBE_MEDIA, data, url)

        elif "video" in data["extension"]:
            media = await self._object_creator(ObjectTypes.WEVERSE_MEDIA, data, url)

        self._log(("MEDIA", "Media"), data, media)
        return media

    async def fetch_live(self, live_id: str) -> Live:
        """Fetches a Weverse Live Broadcast by its ID.

        Parameters
        ----------
        live_id: :class:`str`
            The ID of the live broadcast to fetch.

        Returns
        -------
        :class:`sunverse.objects.Live`
            The Live object requested by the user.
        """
        url = self._urls.post_url(live_id)
        data = await self._fetcher(url)
        live = await self._object_creator(ObjectTypes.LIVE, data, url)
        self._log(("LIVE", "Live"), data, live)
        return live

    async def fetch_moment(self, moment_id: str) -> Moment | OldMoment:
        """Fetches a Weverse Moment by its ID.

        Parameters
        ----------
        moment_id: :class:`str`
            The ID of the moment to fetch.

        Returns
        -------
        Union[:class:`sunverse.objects.Moment`, :class:`sunverse.objects.OldMoment`]
            The Moment object requested by the user.
        """
        url = self._urls.post_url(moment_id)
        data = await self._fetcher(url)

        if "moment" in data["extension"]:
            moment = await self._object_creator(ObjectTypes.NEW_MOMENT, data, url)

        elif "momentW1" in data["extension"]:
            moment = await self._object_creator(ObjectTypes.OLD_MOMENT, data, url)

        self._log(("MOMENT", "Moment"), data, moment)
        return moment

    async def fetch_notice(self, notice_id: int) -> Notice:
        """Fetches a Weverse Notice by its ID.

        Parameters
        ----------
        notice_id: :class:`int`
            The ID of the notice to fetch.

        Returns
        -------
        :class:`sunverse.objects.Notice`
            The Notice object requested by the user.
        """
        url = self._urls.notice_url(notice_id)
        data = await self._fetcher(url)
        notice = await self._object_creator(ObjectTypes.NOTICE, data, url)
        self._log(("NOTICE", "Notice"), data, notice)
        return notice

    async def fetch_member(self, member_id: str) -> Member:
        """Fetches a Weverse Member by its ID.

        Parameters
        ----------
        member_id: :class:`str`
            The ID of the member to fetch.

        Returns
        -------
        :class:`sunverse.objects.Member`
            The Member object requested by the user.
        """
        url = self._urls.member_url(member_id)
        data = await self._fetcher(url)
        member = await self._object_creator(ObjectTypes.MEMBER, data, url)
        self._log(("MEMBER", "Member"), data, member)
        return member

    async def fetch_artist_comments(self, post_id: str) -> list[Comment]:
        """Fetches a list of artist comments that belong to a
        post by the post ID.

        Parameters
        ----------
        post_id: :class:`str`
            The ID of the post that the artist comments are part of to fetch.

        Returns
        -------
        list[:class:`sunverse.objects.Comment`]
            The list that contains the Comment objects requested by the user.
        """
        artist_comments = []
        url = self._urls.artist_comments_url(post_id)
        data = await self._fetcher(url)

        for comment_data in data["data"]:
            artist_comment = await self._object_creator(
                ObjectTypes.COMMENT, comment_data, url
            )
            artist_comments.append(artist_comment)

        self._log(("ARTIST COMMENTS", "Artist Comments"), data, artist_comments)
        return artist_comments

    async def fetch_comment(self, comment_id: str) -> Comment:
        """Fetches a Weverse Comment by its ID.

        Parameters
        ----------
        comment_id: :class:`str`
            The ID of the comment to fetch.

        Returns
        -------
        :class:`sunverse.objects.Comment`
            The Comment object requested by the user.
        """
        url = self._urls.comment_url(comment_id)
        data = await self._fetcher(url)
        comment = await self._object_creator(ObjectTypes.COMMENT, data, url)
        self._log(("COMMENT", "Comment"), data, comment)
        return comment

    async def start(self) -> asyncio.Task:
        """Starts the Sunverse Client."""
        await self._fetch_access_token_by_credentials()
        await self.fetch_latest_notifications()
        logger.info(
            "Started listening for Weverse Notifications. "
            "You will be updated every %s seconds.",
            self.stream_delay,
        )
        self.__task = asyncio.create_task(self.__notification_stream())
        return self.__task

    def stop(self) -> None:
        """Disconnects the Sunverse Client."""
        self.__task.cancel()
        self.__task = None
        logger.info("The Weverse client has been stopped.")

    async def __notification_stream(self) -> None:
        while True:
            try:
                await asyncio.sleep(self.stream_delay)
                notifications, comments = await self.fetch_new_notification()

                for notification in notifications:
                    await self.on_new_notification(notification)

                    if notification.post_type == PostTypes.POST:
                        post = await self.fetch_post(notification.post_id)
                        await self.on_new_post(post)

                    elif notification.post_type == PostTypes.MOMENT:
                        moment = await self.fetch_moment(notification.post_id)
                        await self.on_new_moment(moment)

                    elif notification.post_type == PostTypes.MEDIA:
                        media = await self.fetch_media(notification.post_id)
                        if media:  # Don't call this method when the media is paid.
                            await self.on_new_media(media)

                    elif notification.post_type == PostTypes.LIVE:
                        live = await self.fetch_live(notification.post_id)
                        await self.on_new_live(live)

                    elif notification.post_type == PostTypes.NOTICE:
                        notice = await self.fetch_notice(notification.post_id)
                        await self.on_new_notice(notice)

                for comment in comments:
                    await self.on_new_comment(comment)

            except Exception as exception:
                await self.on_exception(exception)

    async def on_new_notification(self, notification: Notification) -> None:
        """This method is called when a new notification is detected.
        You can overwrite this method to do what you want with the new
        notification.
        """

    async def on_new_post(self, post: Post) -> None:
        """This method is called when a new post is detected.
        You can overwrite this method to do what you want with the new post.
        """

    async def on_new_moment(self, moment: Moment | OldMoment) -> None:
        """This method is called when a new moment is detected.
        You can overwrite this method to do what you want with the new moment.
        """

    async def on_new_media(
        self, media: ImageMedia | WeverseMedia | YoutubeMedia
    ) -> None:
        """This method is called when a new media is detected.
        You can overwrite this method to do what you want with the new media.
        """

    async def on_new_live(self, live: Live) -> None:
        """This method is called when a new live is detected.
        You can overwrite this method to do what you want with the new live.
        """

    async def on_new_notice(self, notice: Notice) -> None:
        """This method is called when a new notice is detected.
        You can overwrite this method to do what you want with the new notice.
        """

    async def on_new_comment(self, comment: Comment) -> None:
        """This method is called when a new comment is detected.
        You can overwrite this method to do what you want with the new comment.
        """

    async def on_exception(self, exception: Exception) -> None:
        """This method is called when an unhandled exception occurs.
        You can overwrite this method to do what you want with the exception.

        By default, an exception message is logged.
        """
        logger.exception(exception)


class _SunverseCache:
    def __init__(self):
        self.comment_noti_cache: dict[str, int] = {}
        self.community_cache: dict[int, Community] = {}
        self.artist_cache: dict[str, Artist] = {}
        self.notification_cache: dict[int, Notification] = {}

    def cache_comment_noti(self, comment_notification: Notification) -> None:
        """Caches comment notifications so that they can be used
        to determine how many and which comment notifications are new.

        Parameters
        ----------
        comment_notification: :class:`sunverse.objects.Notification`
            The comment notification to cache.
        """
        self.comment_noti_cache[
            comment_notification.post_id + comment_notification.author_id
        ] = comment_notification.count

    def limitless_cache(self, object_type: str, sunverse_object) -> None:
        """Responsible for caching objects that are not limited by
        the defined limit. See the notes below for more details
        regarding which objects are cached using this method.

        Parameters
        ----------
        object_type: :class:`str`
            The type of object to cache.
        sunverse_object: :class:`Any`
            The actual Sunverse Object to cache.

        Notes
        -----
        Limitless caching is applied to :class:`sunverse.objects.Community`,
        :class:`sunverse.objects.Artist` and :class:`sunverse.objects.Notification`.
        """
        cache_mapping = {
            ObjectTypes.COMMUNITY: self.community_cache,
            ObjectTypes.ARTIST: self.artist_cache,
            ObjectTypes.NOTIFICATION: self.notification_cache,
        }
        if object_type == ObjectTypes.ARTIST:
            for artist in sunverse_object:
                cache_mapping[object_type][artist] = sunverse_object[artist]

        else:
            cache_mapping[object_type][sunverse_object.id] = sunverse_object
